# Copyright 2018 Jetperch LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Test stream buffer
"""

import unittest
import numpy as np
import pyximport; pyximport.install(setup_args={'include_dirs': np.get_include()})
from joulescope.v0.stream_buffer import Statistics, RawProcessor, StreamBuffer, \
    usb_packet_factory, STATS_DTYPE, STATS_FIELD_COUNT

SAMPLES_PER = 126


# from https://docs.scipy.org/doc/numpy/user/basics.rec.html
def _print_offsets(d):
    print("offsets:", [d.fields[name][1] for name in d.names])
    print("itemsize:", d.itemsize)


def _init(data):
    d = np.zeros(STATS_FIELD_COUNT, dtype=STATS_DTYPE)
    for k in range(len(d)):
        for j in range(len(d[k])):
            d[k][j] = data[k][j]
    return d


def single_stat_as_array(a):
    try:
        a = [a[name] for name in a.dtype.names]
    except Exception:
        pass
    return a


def assert_single_stat_close(a, b):
    a = single_stat_as_array(a)
    b = single_stat_as_array(b)
    return np.testing.assert_allclose(a, b)


def assert_stat_close(a, b):
    return [assert_single_stat_close(x, y) for x, y in zip(a, b)]


class TestStatistics(unittest.TestCase):

    def test_initialize_empty(self):
        s = Statistics()
        self.assertEqual(0, len(s))
        fields = ('length', 'mean', 'variance', 'min', 'max')
        self.assertEqual(fields, s.value.dtype.names)
        _print_offsets(s.value.dtype)
        np.testing.assert_allclose(0, s.value[0]['length'])
        np.testing.assert_allclose(0, s.value[0]['mean'])
        np.testing.assert_allclose(0, s.value[0]['variance'])

    def test_initialize_zero(self):
        d = np.zeros(STATS_FIELD_COUNT, dtype=STATS_DTYPE)
        s = Statistics(stats=d)
        self.assertEqual(0, len(s))
        assert_stat_close(d, s.value)

    def test_combine(self):
        d1 = _init([[1, 1, 0, 1, 1], [1, 2, 0, 2, 2], [1, 3, 0, 3, 3], [1, 4, 0, 4, 4], [1, 0, 0, 0, 0], [1, 1, 0, 1, 1]])
        d2 = _init([[1, 3, 0, 3, 3], [1, 4, 0, 4, 4], [1, 5, 0, 5, 5], [1, 6, 0, 6, 6], [1, 1, 0, 1, 1], [1, 0, 0, 0, 0]])
        e = _init([[2, 2, 2, 1, 3], [2, 3, 2, 2, 4], [2, 4, 2, 3, 5], [2, 5, 2, 4, 6], [2, 0.5, 0.5, 0, 1], [2, 0.5, 0.5, 0, 1]])
        s1 = Statistics(stats=d1)
        s2 = Statistics(stats=d2)
        s1.combine(s2)
        self.assertEqual(2, len(s1))
        assert_stat_close(e, s1.value)

    def test_combine_other_empty(self):
        d = _init([[10, 1, 0, 1, 1], [10, 2, 0, 2, 2], [10, 3, 0, 3, 3], [10, 4, 0, 4, 4], [10, 0, 0, 0, 0], [10, 1, 0, 1, 1]])
        s1 = Statistics(stats=d)
        s2 = Statistics()
        s1.combine(s2)
        self.assertEqual(10, len(s1))
        assert_stat_close(d, s1.value)

    def test_combine_self_empty(self):
        d = _init([[10, 1, 0, 1, 1], [10, 2, 0, 2, 2], [10, 3, 0, 3, 3], [10, 4, 0, 4, 4], [10, 0, 0, 0, 0], [10, 1, 0, 1, 1]])
        s1 = Statistics()
        s2 = Statistics(stats=d)
        s1.combine(s2)
        self.assertEqual(10, len(s1))
        assert_stat_close(d, s1.value)

    def test_combine_both_empty(self):
        s1 = Statistics()
        s2 = Statistics()
        s1.combine(s2)
        self.assertEqual(0, len(s1))


class TestRawProcessor(unittest.TestCase):

    def test_bulk(self):
        r = RawProcessor()
        r.suppress_mode = 'off'
        self.assertEqual('off', r.suppress_mode)
        r.calibration_set([-10.0] * 7, [2.0] * 7, [-2.0, 1.0], [4.0, 1.0])
        raw = usb_packet_factory(0, 1)[8:].view(dtype=np.uint16)
        cal, bits = r.process_bulk(raw)
        self.assertEqual((126, 2), cal.shape)
        np.testing.assert_allclose([[-20.0, -4.0], [-16, 4], [-12, 12]], cal[0:3, :])
        np.testing.assert_allclose([0x20, 0x20, 0x20], bits[0:3])

    def test_bulk_skip(self):
        r = RawProcessor()
        r.suppress_mode = 'off_0_0_0'
        r.calibration_set([-10.0] * 7, [2.0] * 7, [-2.0, 1.0], [4.0, 1.0])
        raw = np.full((20, ), 0xffff, dtype=np.uint16)
        cal, bits = r.process_bulk(raw)
        np.testing.assert_allclose(8, np.bitwise_and(0x0f, bits))

    def generate(self, mode, i_range=None, range_idx=None):
        range_idx = 16 if range_idx is None else int(range_idx)
        if i_range is None:
            i_range = 0, 1
        length = max(32, range_idx + 16)
        r = RawProcessor()
        r.suppress_mode = mode
        r.calibration_set([-1000] * 7, [0.1 ** x for x in range(7)],
                          [-1000, -1000], [0.001, 0.0001])
        current = np.zeros(length, dtype=np.int16)
        current[:] = 2000
        voltage = np.zeros(length, dtype=np.int16)
        voltage[:] = 2000

        raw = np.empty(length * 2, dtype=np.uint16)
        i_range_ = np.zeros(length, dtype=np.uint8)
        i_range_[:range_idx] = i_range[0]
        i_range_[range_idx:] = i_range[1]
        raw[::2] = np.bitwise_or(
            np.left_shift(current, 2),
            np.bitwise_and(i_range_, 0x03))
        raw[1::2] = np.bitwise_or(
            np.left_shift(voltage, 2),
            np.bitwise_and(np.right_shift(i_range_, 2), 0x03))
        raw[3::4] = np.bitwise_or(raw[3::4], 0x0002)
        return r.process_bulk(raw)

    def test_invalid_mode(self):
        r = RawProcessor()
        with self.assertRaises(ValueError):
            r.suppress_mode = 'mean_0_'

    def test_malformed(self):
        r = RawProcessor()
        with self.assertRaises(ValueError):
            r.suppress_mode = 'mean_0_0'

    def test_off(self):
        for k in range(6):
            with self.subTest(i=k):
                cal, _ = self.generate('off', i_range=(k, k + 1))
                g = 0.1 ** k
                np.testing.assert_allclose(g * 1000, cal[10:16, 0])
                np.testing.assert_allclose(g * 100, cal[16:, 0])
                np.testing.assert_allclose(1, cal[:, 1])

    def test_off_1_2_1(self):
        for k in range(6):
            with self.subTest(i=k):
                cal, _ = self.generate('off_1_2_1', i_range=(k, k + 1))
                g = 0.1 ** k
                np.testing.assert_allclose(g * 1000, cal[10:16, 0])
                np.testing.assert_allclose(g * 100, cal[16:, 0])
                np.testing.assert_allclose(1, cal[:, 1])

    def test_nan_0_1_0(self):
        cal, _ = self.generate('nan_0_1_0')
        np.testing.assert_allclose(1000, cal[10:16, 0])
        self.assertFalse(np.isfinite(cal[16, 0]))
        np.testing.assert_allclose(100, cal[17:, 0])

    def test_nan_0_2_0(self):
        cal, _ = self.generate('nan_0_2_0')
        np.testing.assert_allclose(1000, cal[10:16, 0])
        self.assertFalse(np.isfinite(cal[16, 0]))
        self.assertFalse(np.isfinite(cal[17, 0]))
        np.testing.assert_allclose(100, cal[18:, 0])

    def test_mean_1_3_1(self):
        cal, _ = self.generate('mean_1_3_1')
        np.testing.assert_allclose(1000, cal[10:16, 0])
        mean_expect = np.mean([cal[15, 0], cal[19, 0]])
        np.testing.assert_allclose(mean_expect, cal[16:19, 0])
        np.testing.assert_allclose(100, cal[19:, 0])

    def test_mean_2_3_1(self):
        cal, _ = self.generate('mean_2_3_1')
        np.testing.assert_allclose(1000, cal[10:16, 0])
        mean_expect = np.mean([cal[14, 0], cal[15, 0], cal[19, 0]])
        np.testing.assert_allclose(mean_expect, cal[16:19, 0])
        np.testing.assert_allclose(100, cal[19:, 0])

    def test_mean_1_3_2(self):
        cal, _ = self.generate('mean_1_3_2')
        np.testing.assert_allclose(1000, cal[10:16, 0])
        mean_expect = np.mean([cal[15, 0], cal[19, 0], cal[20, 0]])
        np.testing.assert_allclose(mean_expect, cal[16:19, 0])
        np.testing.assert_allclose(100, cal[19:, 0])

    def test_mean_1_m_1(self):
        for k in range(6):
            with self.subTest(i=k):
                cal, _ = self.generate('mean_1_m_1', i_range=(k, k + 1))
                n = 3 if k == 0 else 4
                z = 16 + n
                g = 0.1 ** k
                np.testing.assert_allclose(g * 1000, cal[10:16, 0])
                np.testing.assert_allclose(g * 550, cal[16:z, 0])
                np.testing.assert_allclose(g * 100, cal[z:, 0])

    def test_mean_1_n_1(self):
        for k in range(6):
            with self.subTest(i=k):
                cal, _ = self.generate('mean_1_n_1', i_range=(k, k + 1))
                n = 3 if k == 0 else 5
                z = 16 + n
                g = 0.1 ** k
                np.testing.assert_allclose(g * 1000, cal[10:16, 0])
                np.testing.assert_allclose(g * 550, cal[16:z, 0])
                np.testing.assert_allclose(g * 100, cal[z:, 0])

    def test_interp_1_7_1(self):
        k = 7
        cal, _ = self.generate(f'interp_1_{k}_1', i_range=(3, 2))
        g = 0.1 ** 3
        idx_end = 15 + k + 2
        v_start = g * 1000
        v_end = g * 10000
        v_interp = np.linspace(v_start, v_end, k + 2)
        np.testing.assert_allclose(v_start, cal[10:15, 0])
        np.testing.assert_allclose(v_interp, cal[15:idx_end, 0])
        np.testing.assert_allclose(v_end, cal[idx_end:, 0])

    def test_suppress_history_rollover_2_1_0(self):
        for k in range(14, 72):
            with self.subTest(i=k):
                cal, _ = self.generate('mean_2_1_0', range_idx=k)
                y = np.mean(cal[(k - 2):k, 0])
                np.testing.assert_allclose(y, cal[k, 0])


class TestStreamBuffer(unittest.TestCase):

    def test_init(self):
        b = StreamBuffer(1.0, [10, 10], 1000.0)
        self.assertIsNotNone(b)
        self.assertEqual(1000, len(b))
        self.assertEqual(1000.0, b.output_sampling_frequency)
        self.assertEqual((0, 0), b.sample_id_range)
        self.assertEqual((0.0, 1.0), b.limits_time)
        self.assertEqual((-1000, 0), b.limits_samples)
        self.assertEqual(0, b.time_to_sample_id(1.0))
        self.assertEqual(1.0, b.sample_id_to_time(0))
        del b

    def test_insert_process(self):
        b = StreamBuffer(1.0, [100, 10], 1000.0)
        b.suppress_mode = 'off'
        frame = usb_packet_factory(0, 1)
        self.assertEqual(0, b.status()['device_sample_id']['value'])
        b.insert(frame)
        self.assertEqual(126, b.status()['device_sample_id']['value'])
        self.assertEqual(0, b.status()['sample_id']['value'])
        b.process()
        self.assertEqual((0, 126), b.sample_id_range)
        self.assertEqual((0.0, 1.0), b.limits_time)
        self.assertEqual((-874, 126), b.limits_samples)
        self.assertEqual(126, b.time_to_sample_id(1.0))
        self.assertEqual(1.0, b.sample_id_to_time(126))
        self.assertEqual(0.999, b.sample_id_to_time(125))
        self.assertEqual(b.limits_time[0], b.sample_id_to_time(b.limits_samples[0]))
        self.assertEqual(b.limits_time[1], b.sample_id_to_time(b.limits_samples[1]))
        self.assertEqual(126, b.status()['sample_id']['value'])
        samples = b.samples_get(0, 126, ['raw'])
        raw_data = b.samples_get(0, 126, 'raw')
        np.testing.assert_allclose(samples['signals']['raw']['value'], raw_data)
        expect = np.arange(126*2, dtype=np.uint16).reshape((126, 2))
        np.testing.assert_allclose(expect, np.right_shift(raw_data, 2))
        np.testing.assert_allclose(expect, b.data_buffer[0:126*2].reshape((126, 2)))
        data = b.data_get(0, 126)
        np.testing.assert_allclose(expect[:, 0], data[:, 0]['mean'])

    def test_wrap_aligned(self):
        frame = usb_packet_factory(0, 4)
        b = StreamBuffer(2.0 * SAMPLES_PER / 1000.0, [], 1000.0)
        b.suppress_mode = 'off'
        b.insert(frame)
        b.process()
        self.assertEqual((SAMPLES_PER * 2, SAMPLES_PER * 4), b.sample_id_range)
        data = b.samples_get(SAMPLES_PER * 2, SAMPLES_PER * 4, 'raw')
        data = np.right_shift(data, 2)
        expect = np.arange(SAMPLES_PER * 4, SAMPLES_PER * 8, dtype=np.uint16).reshape((SAMPLES_PER * 2, 2))
        np.testing.assert_allclose(expect, data)
        np.testing.assert_allclose(expect[:, 0], b.data_buffer[::2])
        data = b.data_get(SAMPLES_PER * 2, SAMPLES_PER * 4)
        np.testing.assert_allclose(expect[:, 0], data[:, 0]['mean'])
        np.testing.assert_allclose(expect[:, 1], data[:, 1]['mean'])

    def test_wrap_unaligned(self):
        frame = usb_packet_factory(0, 4)
        b = StreamBuffer((2 * SAMPLES_PER + 2) / 1000.0, [], 1000.0)
        b.insert(frame)
        b.process()
        self.assertEqual(SAMPLES_PER * 4, b.sample_id_range[1])
        data = np.right_shift(b.samples_get(SAMPLES_PER * 2, SAMPLES_PER * 4, 'raw'), 2)
        expect = np.arange(SAMPLES_PER * 4, SAMPLES_PER * 8, dtype=np.uint16).reshape((SAMPLES_PER * 2, 2))
        np.testing.assert_allclose(expect, data)

    def test_get_over_samples(self):
        b = StreamBuffer(2.0, [10, 10], 1000.0)
        b.suppress_mode = 'off'
        frame = usb_packet_factory(0, 1)
        b.insert(frame)
        b.process()
        data = b.data_get(0, 21, 5)
        self.assertEqual((4, STATS_FIELD_COUNT), data.shape)
        np.testing.assert_allclose(np.arange(10), b.data_buffer[0:10])  # processed correctly
        np.testing.assert_allclose([4.0, 14.0, 24.0, 34.0], data[:, 0]['mean'])
        np.testing.assert_allclose([5, 4.0, 40.0, 0.0, 8.0], single_stat_as_array(data[0, 0]))
        np.testing.assert_allclose([5.0, 15.0, 25.0, 35.0], data[:, 1]['mean'])

    def test_get_over_reduction_direct(self):
        b = StreamBuffer(2.0, [10, 10], 1000.0)
        b.suppress_mode = 'off'
        frame = usb_packet_factory(0, 1)
        b.insert(frame)
        b.process()
        data = b.data_get(0, 20, 10)
        self.assertEqual((2, STATS_FIELD_COUNT), data.shape)
        np.testing.assert_allclose([10, 9.0, 330.0, 0.0, 18.0], single_stat_as_array(data[0, 0]))
        np.testing.assert_allclose([10, 29.0, 330.0, 20.0, 38.0], single_stat_as_array(data[1, 0]))
        np.testing.assert_allclose([10, 10.0, 330.0, 1.0, 19.0], single_stat_as_array(data[0, 1]))
        np.testing.assert_allclose([10, 30.0, 330.0, 21.0, 39.0], single_stat_as_array(data[1, 1]))

    def test_get_over_reduction(self):
        b = StreamBuffer(2.0, [10, 10], 1000.0)
        b.suppress_mode = 'off'
        frame = usb_packet_factory(0, 1)
        b.insert(frame)
        b.process()
        data = b.data_get(0, 40, 20)
        self.assertEqual((2, STATS_FIELD_COUNT), data.shape)
        assert_single_stat_close([20, 19.0, 2660.0, 0.0, 38.0], data[0, 0])
        assert_single_stat_close([20, 20.0, 2660.0, 1.0, 39.0], data[0, 1])
        assert_single_stat_close([20, 59.0, 2660.0, 40.0, 78.0], data[1, 0])

    def test_get_over_reduction_direct_with_raw_nan(self):
        b = StreamBuffer(2.0, [10, 10], 1000.0)
        frame = usb_packet_factory(0, 1)
        frame[8+11*4:8+15*4:] = 0xff
        b.insert(frame)
        b.process()
        r = b.get_reduction(0, 0, 126)
        self.assertTrue(all(np.isfinite(r[:, 0]['mean'])))

    def test_get_over_reduction_direct_with_reduction0_nan_mean(self):
        b = StreamBuffer(2.0, [10, 10], 1000.0)
        b.suppress_mode = 'mean_1_n_1'
        frame = usb_packet_factory(0, 1)
        frame[8+10*4:8+22*4:] = 0xff
        b.insert(frame)
        b.process()
        r0 = b.get_reduction(0, 0, 126)
        self.assertFalse(np.isfinite(r0[1, 0]['mean']))
        r1 = b.get_reduction(1, 0, 126)
        self.assertTrue(np.isfinite(r1[0, 0]['mean']))

    def test_get_over_reduction_direct_with_reduction0_nan_interp(self):
        b = StreamBuffer(2.0, [10, 10], 1000.0)
        self.assertEqual('interp_1_n_1', b.suppress_mode)
        frame = usb_packet_factory(0, 1)
        frame[8+10*4:8+22*4:] = 0xff
        b.insert(frame)
        b.process()
        r0 = b.get_reduction(0, 0, 126)
        self.assertFalse(np.isfinite(r0[1, 0]['mean']))
        r1 = b.get_reduction(1, 0, 126)
        self.assertTrue(np.isfinite(r1[0, 0]['mean']))

    def test_calibration(self):
        b = StreamBuffer(2.0, [10, 10], 1000.0)
        b.suppress_mode = 'off'
        self.assertEqual('off', b.suppress_mode)
        b.calibration_set([-10.0] * 7, [2.0] * 7, [-2.0, 1.0], [4.0, 1.0])
        frame = usb_packet_factory(0, 1)
        b.insert(frame)
        b.process()
        data = b.data_get(0, 10, 1)
        self.assertEqual((10, STATS_FIELD_COUNT), data.shape)
        np.testing.assert_allclose([-20.0, -4.0, 80.0, 0, 0, 1], data[0, :]['mean'])
        np.testing.assert_allclose([12.,  60., 720., 0, 0, 1], data[8, :]['mean'])

    def stream_buffer_01(self):
        b = StreamBuffer(2.0, [10, 10], 1000.0)
        b.suppress_mode = 'off'
        frame = usb_packet_factory(0, 1)
        b.insert(frame)
        b.process()
        return b

    def stream_buffer_02(self):
        b = StreamBuffer(2.0, [10, 10], 1000.0)
        b.suppress_mode = 'off'
        b.insert(usb_packet_factory(0))
        b.insert(usb_packet_factory(2))
        b.process()
        return b

    def test_stats_direct(self):
        b = self.stream_buffer_01()
        s = b.statistics_get(13, 13)[0]
        self.assertEqual(0, s[0]['length'])
        np.testing.assert_allclose(b.data_get(13, 14, 1)[0, 0]['mean'], b.statistics_get(13, 14)[0][0]['mean'])
        np.testing.assert_allclose(np.mean(b.data_get(13, 29, 1)[:, 0]['mean']), b.statistics_get(13, 29)[0][0]['mean'])
        self.assertEqual(0, b.status()['skip_count']['value'])

    def test_stats_direct_nan(self):
        b = self.stream_buffer_02()
        self.assertEqual((0, 126 * 3), b.sample_id_range)
        d = b.data_get(0, 126 * 3, 1)[:, 0]['mean']
        self.assertTrue(np.all(np.logical_not(np.isnan(d[0:126]))))
        self.assertTrue(np.all(np.isnan(d[126:252])))
        self.assertTrue(np.all(np.logical_not(np.isnan(d[252:]))))

        self.assertFalse(np.isnan(b.statistics_get(121, 126)[0][0]['mean']))
        self.assertFalse(np.isnan(b.statistics_get(124, 129)[0][0]['mean']))
        self.assertTrue(np.isnan(b.statistics_get(130, 135)[0][0]['mean']))
        self.assertTrue(np.isnan(b.statistics_get(247, 252)[0][0]['mean']))
        self.assertFalse(np.isnan(b.statistics_get(249, 254)[0][0]['mean']))
        self.assertEqual(1, b.status()['skip_count']['value'])
        self.assertEqual(126, b.status()['sample_missing_count']['value'])

    def test_stats_over_single_reduction_exact(self):
        b = self.stream_buffer_01()
        np.testing.assert_allclose(np.mean(b.data_get(10, 20, 1)[:, 0]['mean']), b.statistics_get(10, 20)[0][0]['mean'])

    def test_stats_over_single_reduction_leading(self):
        b = self.stream_buffer_01()
        np.testing.assert_allclose(np.mean(b.data_get(9, 20, 1)[:, 0]['mean']), b.statistics_get(9, 20)[0][0]['mean'])
        np.testing.assert_allclose(np.mean(b.data_get(8, 20, 1)[:, 0]['mean']), b.statistics_get(8, 20)[0][0]['mean'])

    def test_stats_over_single_reduction_trailing(self):
        b = self.stream_buffer_01()
        np.testing.assert_allclose(np.mean(b.data_get(10, 21, 1)[:, 0]['mean']), b.statistics_get(10, 21)[0][0]['mean'])
        np.testing.assert_allclose(np.mean(b.data_get(10, 22, 1)[:, 0]['mean']), b.statistics_get(10, 22)[0][0]['mean'])

    def test_stats_over_single_reduction_extended(self):
        b = self.stream_buffer_01()
        np.testing.assert_allclose(np.mean(b.data_get(9, 21, 1)[:, 0]['mean']), b.statistics_get(9, 21)[0][0]['mean'])
        np.testing.assert_allclose(np.mean(b.data_get(5, 25, 1)[:, 0]['mean']), b.statistics_get(5, 25)[0][0]['mean'])

    def test_stats_over_single_reductions(self):
        b = self.stream_buffer_01()
        np.testing.assert_allclose(np.mean(b.data_get(9, 20, 1)[:, 0]['mean']), b.statistics_get(9, 20)[0][0]['mean'])
        np.testing.assert_allclose(np.mean(b.data_get(10, 21, 1)[:, 0]['mean']), b.statistics_get(10, 21)[0][0]['mean'])
        np.testing.assert_allclose(np.mean(b.data_get(9, 21, 1)[:, 0]['mean']), b.statistics_get(9, 21)[0][0]['mean'])
        np.testing.assert_allclose(np.mean(b.data_get(9, 101, 1)[:, 0]['mean']), b.statistics_get(9, 101)[0][0]['mean'])
        np.testing.assert_allclose(np.mean(b.data_get(5, 105, 1)[:, 0]['mean']), b.statistics_get(5, 105)[0][0]['mean'])
        self.assertEqual(0, b.status()['skip_count']['value'])

    def test_stats_over_single_reduction_nan(self):
        b = self.stream_buffer_02()
        np.testing.assert_allclose(np.mean(b.data_get(120, 126, 1)[:, 0]['mean']), b.statistics_get(120, 130)[0][0]['mean'])
        np.testing.assert_allclose(np.mean(b.data_get(120, 126, 1)[:, 0]['mean']), b.statistics_get(120, 140)[0][0]['mean'])
        np.testing.assert_allclose(np.mean(b.data_get(110, 126, 1)[:, 0]['mean']), b.statistics_get(110, 130)[0][0]['mean'])
        self.assertEqual(1, b.status()['skip_count']['value'])
        self.assertEqual(126, b.status()['sample_missing_count']['value'])

    def test_stats_over_reductions_nan(self):
        b = self.stream_buffer_02()
        np.testing.assert_allclose(np.mean(b.data_get(0, 126, 1)[:, 0]['mean']), b.statistics_get(0, 200)[0][0]['mean'])
        self.assertEqual(1, b.status()['skip_count']['value'])
        self.assertEqual(126, b.status()['sample_missing_count']['value'])

    def test_insert_raw_simple(self):
        b = StreamBuffer(1.0, [100, 100, 100], 1000.0)
        b.suppress_mode = 'off'
        expect = np.arange(126 * 2, dtype=np.uint16).reshape((126, 2))
        raw = np.left_shift(expect, 2)
        raw[1::2, 1] = np.bitwise_or(raw[1::2, 1], 0x0002)
        b.insert_raw(raw)
        b.process()
        data = b.data_get(0, 126)
        np.testing.assert_allclose(expect[:, 0], data[:, 0]['mean'])
        self.assertEqual(0, b.status()['skip_count']['value'])

    def test_insert_raw_wrap(self):
        b = StreamBuffer(0.2, [], 1000.0)
        expect = np.arange(250 * 2, dtype=np.uint16).reshape((250, 2))
        raw = np.left_shift(expect, 2)
        raw[1::2, 1] = np.bitwise_or(raw[1::2, 1], 0x0002)
        b.insert_raw(raw[:100])
        b.process()
        b.insert_raw(raw[100:])
        b.process()
        data = b.data_get(50, 250)
        np.testing.assert_allclose(expect[50:, 0], data[:, 0]['mean'])
        self.assertEqual(0, b.status()['skip_count']['value'])

    def test_insert_former_nan_case(self):
        raw = np.array([
            [61957, 16937],
            [62317, 16935],
            [62585, 16937],
            [62853, 16935],
            [65535, 16916],  # raw_i = 0xffff  (i_range=3)
            [18932, 16942],
            [8876, 16932],
            [9788, 16938],
            [10300, 16936],
            [10368, 16930],
            [10528, 16932],
            [10584, 16938],
            [10564, 16936],
            [10568, 16942],
            [12497, 16932],
            [12733, 16946],
            [12613, 16940],
            [12561, 16930]], dtype=np.uint16)
        b = StreamBuffer(0.2, [], 1000.0)
        b.insert_raw(raw)
        b.process()
        self.assertEqual(0, b.status()['skip_count']['value'])

    def test_i_range_off(self):
        raw = np.array([
            [0x1003, 0x1001],
            [0x1003, 0x1003],
            [0x1003, 0x1001],
            [0x1003, 0x1003],
            [0x1003, 0x1001],
            [0x1003, 0x1003],
            [0x1003, 0x1001],
            [0x1003, 0x1003]], dtype=np.uint16)
        b = StreamBuffer(0.2, [], 1000.0)
        b.insert_raw(raw)
        b.process()
        np.testing.assert_allclose(0, b.samples_get(0, 8, fields='current'))

    def test_voltage_range(self):
        b = StreamBuffer(0.2, [], 1000.0)
        self.assertEqual(0, b.voltage_range)
        b.voltage_range = 1
        self.assertEqual(1, b.voltage_range)



# Copyright 2018 Jetperch LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from joulescope.v0.usb import core as usb_core
from joulescope.v0.usb.api import DeviceEvent
from joulescope.v0.usb.impl_tools import RunUntilDone
from joulescope.v0.usb.core import SetupPacket, ControlTransferResponse
from joulescope.v0.usb.scan_info import INFO
from typing import List
import time
import threading
from contextlib import contextmanager
import platform
import numpy as np
import os
import sys
import struct
import ctypes
import ctypes.util
from ctypes import Structure, c_uint8, c_uint16, c_uint32, c_uint, \
    c_int, c_char, c_ssize_t, c_void_p, POINTER, pointer, byref
import logging

log = logging.getLogger(__name__)
STRING_LENGTH_MAX = 255
TRANSFER_TIMEOUT_MS = 1000  # default in milliseconds


if platform.system() == 'Darwin' and getattr(sys, 'frozen', False):
    machine = platform.machine()
    os_version = platform.release().split('.')[0]
    find_lib = os.path.join(sys._MEIPASS, f'{machine}_{os_version}_libusb-1.0.0.dylib')
    log.info('Darwin lib: %s', find_lib)
else:
    find_lib = ctypes.util.find_library('usb-1.0')
    if find_lib is None:
        raise RuntimeError('Could not import libusb')
_lib = ctypes.cdll.LoadLibrary(find_lib)


class DescriptorType:
    DEVICE = 0x01
    CONFIG = 0x02
    STRING = 0x03
    INTERFACE = 0x04
    ENDPOINT = 0x05
    BOS = 0x0f
    DEVICE_CAPABILITY = 0x10
    HID = 0x21
    REPORT = 0x22
    PHYSICAL = 0x23
    HUB = 0x29
    SUPERSPEED_HUB = 0x2a
    SS_ENDPOINT_COMPANION = 0x30


class TransferType:
    CONTROL = 0
    ISOCHRONOUS = 1
    BULK = 2
    INTERRUPT = 3
    BULK_STREAM = 4


class TransferStatus:
    COMPLETED = 0
    ERROR = 1
    TIMED_OUT = 2
    CANCELLED = 3
    STALL = 4
    NO_DEVICE = 5
    OVERFLOW = 6


class TransferFlags:
    SHORT_NOT_OK = 1 << 0
    FREE_BUFFER = 1 << 1
    FREE_TRANSFER = 1 << 2
    ADD_ZERO_PACKET = 1 << 3


class ReturnCodes:
    SUCCESS = 0
    ERROR_IO = -1
    ERROR_INVALID_PARAM = -2
    ERROR_ACCESS = -3
    ERROR_NO_DEVICE = -4
    ERROR_NOT_FOUND = -5
    ERROR_BUSY = -6
    ERROR_TIMEOUT = -7
    ERROR_OVERFLOW = -8
    ERROR_PIPE = -9
    ERROR_INTERRUPTED = -10
    ERROR_NO_MEM = -11
    ERROR_NOT_SUPPORTED = -12
    ERROR_OTHER = -99


class _libusb_device_descriptor(Structure):
    _fields_ = [
        ('bLength', c_uint8),
        ('bDescriptorType', c_uint8),
        ('bcdUSB', c_uint16),
        ('bDeviceClass', c_uint8),
        ('bDeviceSubClass', c_uint8),
        ('bDeviceProtocol', c_uint8),
        ('bMaxPacketSize0', c_uint8),
        ('idVendor', c_uint16),
        ('idProduct', c_uint16),
        ('bcdDevice', c_uint16),
        ('iManufacturer', c_uint8),
        ('iProduct', c_uint8),
        ('iSerialNumber', c_uint8),
        ('bNumConfigurations', c_uint8)]


# typedef void (LIBUSB_CALL *libusb_transfer_cb_fn)(struct libusb_transfer *transfer);
libusb_transfer_cb_fn = ctypes.CFUNCTYPE(None, c_void_p)

class _libusb_transfer(Structure):
    _fields_ = [
        ('dev_handle', c_void_p),
        ('flags', c_uint8),
        ('endpoint_id', c_uint8),
        ('endpoint_type', c_uint8),
        ('timeout_ms', c_uint),
        ('status', c_int),  # enum libusb_transfer_status
        ('length', c_int),
        ('actual_length', c_int),
        ('callback', libusb_transfer_cb_fn),
        ('user_data', c_void_p),
        ('buffer', POINTER(c_uint8)),
        ('num_iso_packets', c_int),
        # struct libusb_iso_packet_descriptor iso_packet_desc[ZERO_SIZED_ARRAY];
    ]

# typedef struct libusb_context libusb_context; - c_void_p
# typedef struct libusb_device libusb_device; - c_void_p
# typedef struct libusb_device_handle libusb_device_handle; c_void_p

# int LIBUSB_CALL libusb_init(libusb_context **ctx);
_lib.libusb_init.restype = c_int
_lib.libusb_init.argtypes = [POINTER(c_void_p)]

# void LIBUSB_CALL libusb_exit(libusb_context *ctx);
_lib.libusb_exit.restype = None
_lib.libusb_exit.argtypes = [c_void_p]

# ssize_t LIBUSB_CALL libusb_get_device_list(libusb_context *ctx,
#	libusb_device ***list);
_lib.libusb_get_device_list.restype = c_ssize_t
_lib.libusb_get_device_list.argtypes = [c_void_p, POINTER(POINTER(POINTER(c_void_p)))]


# void LIBUSB_CALL libusb_free_device_list(libusb_device **list,
#	int unref_devices);
_lib.libusb_free_device_list.restype = None
_lib.libusb_free_device_list.argtypes = [POINTER(POINTER(c_void_p)), c_int]

# libusb_device * libusb_ref_device (libusb_device *dev)
_lib.libusb_ref_device.restype = POINTER(c_void_p)
_lib.libusb_ref_device.argtypes = [POINTER(c_void_p)]

# void 	libusb_unref_device (libusb_device *dev)
_lib.libusb_unref_device.restype = None
_lib.libusb_unref_device.argtypes = [POINTER(c_void_p)]

# int LIBUSB_CALL libusb_open(libusb_device *dev, libusb_device_handle **dev_handle);
_lib.libusb_open.restype = c_int
_lib.libusb_open.argtypes = [c_void_p, POINTER(c_void_p)]

# void LIBUSB_CALL libusb_close(libusb_device_handle *dev_handle);
_lib.libusb_close.restype = None
_lib.libusb_close.argtypes = [c_void_p]

# int LIBUSB_CALL libusb_set_configuration(libusb_device_handle *dev_handle,
#	int configuration);
_lib.libusb_set_configuration.restype = c_int
_lib.libusb_set_configuration.argtypes = [c_void_p, c_int]

# int LIBUSB_CALL libusb_claim_interface(libusb_device_handle *dev_handle,
#	int interface_number);
_lib.libusb_claim_interface.restype = c_int
_lib.libusb_claim_interface.argtypes = [c_void_p, c_int]

# int LIBUSB_CALL libusb_release_interface(libusb_device_handle *dev_handle,
#	int interface_number);
_lib.libusb_release_interface.restype = c_int
_lib.libusb_release_interface.argtypes = [c_void_p, c_int]

# int LIBUSB_CALL libusb_set_interface_alt_setting(libusb_device_handle *dev_handle,
#	int interface_number, int alternate_setting);
_lib.libusb_set_interface_alt_setting.restype = c_int
_lib.libusb_set_interface_alt_setting.argtypes = [c_void_p, c_int, c_int]


# int LIBUSB_CALL libusb_get_device_descriptor(libusb_device *dev,
#	struct libusb_device_descriptor *desc);
_lib.libusb_get_device_descriptor.restype = c_int
_lib.libusb_get_device_descriptor.argtypes = [c_void_p, POINTER(_libusb_device_descriptor)]

# int LIBUSB_CALL libusb_control_transfer(libusb_device_handle *dev_handle,
#	uint8_t request_type, uint8_t bRequest, uint16_t wValue, uint16_t wIndex,
#	unsigned char *data, uint16_t wLength, unsigned int timeout);
_lib.libusb_control_transfer.restype = c_int
_lib.libusb_control_transfer.argtypes = [c_void_p, c_uint8, c_uint8, c_uint16, c_uint16,
                                         POINTER(c_uint8), c_uint16, c_int]

# struct libusb_transfer * LIBUSB_CALL libusb_alloc_transfer(int iso_packets);
_lib.libusb_alloc_transfer.restype = POINTER(_libusb_transfer)
_lib.libusb_alloc_transfer.argtypes = [c_int]

# int LIBUSB_CALL libusb_submit_transfer(struct libusb_transfer *transfer);
_lib.libusb_submit_transfer.restype = c_int
_lib.libusb_submit_transfer.argtypes = [POINTER(_libusb_transfer)]

# int LIBUSB_CALL libusb_cancel_transfer(struct libusb_transfer *transfer);
_lib.libusb_cancel_transfer.restype = c_int
_lib.libusb_cancel_transfer.argtypes = [POINTER(_libusb_transfer)]

# void LIBUSB_CALL libusb_free_transfer(struct libusb_transfer *transfer);
_lib.libusb_free_transfer.restype = None
_lib.libusb_free_transfer.argtypes = [POINTER(_libusb_transfer)]


class TimeVal(Structure):
    _fields_ = [
        ("tv_sec", ctypes.c_long),
        ("tv_usec", ctypes.c_long)
    ]

# int LIBUSB_CALL libusb_handle_events_timeout(libusb_context *ctx,
#	struct timeval *tv);
_lib.libusb_handle_events_timeout.restype = c_int
_lib.libusb_handle_events_timeout.argtypes = [c_void_p, POINTER(TimeVal)]

# int LIBUSB_CALL libusb_handle_events(libusb_context *ctx)
_lib.libusb_handle_events.restype = c_int
_lib.libusb_handle_events.argtypes = [c_void_p]


class HotplugFlag:
    NONE = 0
    ENUMERATE = 1 << 0


class HotplugEvent:
    DEVICE_ARRIVED = 0x01
    DEVICE_LEFT    = 0x02


HOTPLUG_MATCH_ANY = -1


# typedef int (LIBUSB_CALL *libusb_hotplug_callback_fn)(libusb_context *ctx,
#						libusb_device *device,
#						libusb_hotplug_event event,
#						void *user_data);
_libusb_hotplug_callback_fn = ctypes.CFUNCTYPE(c_int, c_void_p, c_void_p, c_int, c_void_p)

# int LIBUSB_CALL libusb_hotplug_register_callback(libusb_context *ctx,
#						libusb_hotplug_event events,
#						libusb_hotplug_flag flags,
#						int vendor_id, int product_id,
#						int dev_class,
#						libusb_hotplug_callback_fn cb_fn,
#						void *user_data,
#						libusb_hotplug_callback_handle *callback_handle);
_lib.libusb_hotplug_register_callback.restype = c_int
_lib.libusb_hotplug_register_callback.argtypes = [c_void_p, c_int, c_int, c_int, c_int, c_int,
                                                  _libusb_hotplug_callback_fn, c_void_p, POINTER(c_int)]

# void LIBUSB_CALL libusb_hotplug_deregister_callback(libusb_context *ctx,
#						libusb_hotplug_callback_handle callback_handle);
_lib.libusb_hotplug_deregister_callback.restype = c_int
_lib.libusb_hotplug_deregister_callback.argtypes = [c_void_p, c_int]


class Capability:
    HAS_CAPABILITY = 0x0000
    HAS_HOTPLUG = 0x0001
    HAS_HID_ACCESS = 0x0100
    SUPPORTS_DETACH_KERNEL_DRIVER = 0x0101


# int LIBUSB_CALL libusb_has_capability(uint32_t capability);
_lib.libusb_has_capability.restype = c_int
_lib.libusb_has_capability.argtypes = [c_uint32]



def _libusb_context_create():
    ctx = c_void_p()

    rc = _lib.libusb_init(pointer(ctx))
    if rc:
        raise RuntimeError('Could not open libusb')
    return ctx


def _libusb_context_destroy(ctx):
    _lib.libusb_exit(ctx)


@contextmanager
def _libusb_context():
    ctx = _libusb_context_create()
    try:
        yield ctx
    finally:
        _libusb_context_destroy(ctx)


def _path_split(path):
    vid, pid, serial_number = path.split('/')
    return int(vid, 16), int(pid, 16), serial_number


def _get_string_descriptor(device, index):
    request_type = usb_core.RequestType(direction='in', type_='standard', recipient='device').u8
    byte_buffer = bytearray(STRING_LENGTH_MAX)
    buffer_type = c_uint8 * STRING_LENGTH_MAX
    buffer = buffer_type.from_buffer(byte_buffer)

    # determine default language
    rv = _lib.libusb_control_transfer(device, request_type, usb_core.Request.GET_DESCRIPTOR,
                                      (DescriptorType.STRING << 8), 0,
                                      buffer, STRING_LENGTH_MAX,
                                      1000)
    if rv < 0:
        raise RuntimeError('control_transfer could not get language: %d' % (rv, ))
    langid = int(byte_buffer[2]) | (int(byte_buffer[3]) << 8)

    rv = _lib.libusb_control_transfer(device, request_type, usb_core.Request.GET_DESCRIPTOR,
                                      (DescriptorType.STRING << 8) | (index & 0xff), langid,
                                      buffer, STRING_LENGTH_MAX,
                                      1000)

    if rv < 0:
        raise RuntimeError('control transfer could not get string descriptor: %d' % (rv, ))
    buffer_len = min(rv, byte_buffer[0])
    # byte 0 is length, byte 1 is string identifier
    return byte_buffer[2:buffer_len].decode('UTF-16-LE')


_transfer_callback_discard_fn = libusb_transfer_cb_fn(lambda x: None)
"""Default null callback that is always safe."""


class Transfer:

    def __init__(self, size):
        try:
            self.size = len(size)  # also serves as list-like duck-typing test
            self.buffer = np.frombuffer(size, dtype=np.uint8)
            log.debug('Transfer: copy buffer %d', self.size)
        except TypeError:
            self.size = size
            self.buffer = np.full(self.size, 0, dtype=np.uint8)
            log.debug('Transfer: create buffer %d', self.size)
        self.transfer = _lib.libusb_alloc_transfer(0)  # type: _libusb_transfer
        self.addr = ctypes.addressof(self.transfer.contents)
        transfer = self.transfer[0]
        self.buffer_ptr = self.buffer.ctypes.data_as(POINTER(c_uint8))
        transfer.buffer = self.buffer_ptr
        transfer.flags = 0
        transfer.length = self.size
        transfer.actual_length = 0
        transfer.user_data = None
        transfer.num_iso_packets = 0
        transfer.status = TransferStatus.COMPLETED
        transfer.timeout_ms = TRANSFER_TIMEOUT_MS  # milliseconds
        transfer.callback = _transfer_callback_discard_fn

    def close(self):
        if self.buffer is not None:
            buffer, self.buffer = self.buffer, None
            _lib.libusb_free_transfer(self.transfer)

    def __del__(self):
        self.close()


class ControlTransferAsync:

    def __init__(self, handle):
        """Manage asynchronous control transfers.

        :param handle: The device handle.
        """
        self._handle = handle
        self._transfer_callback_fn = libusb_transfer_cb_fn(self._transfer_callback)
        self._commands = []  # Pending control transfer commands as list of [cbk_fn, setup_packet, buffer]
        self._transfer_pending = None  # type: Transfer
        self._time_start = None
        self.stop_code = None

    def __str__(self):
        return 'ControlTransferAsync()'

    def __len__(self):
        return len(self._commands)

    @property
    def is_busy(self):
        return 0 != len(self._commands)

    def _transfer_callback(self, transfer_void_ptr):
        transfer, self._transfer_pending = self._transfer_pending, None
        if transfer is None:
            log.warning('Transfer callback when none pending')
            return
        if transfer.addr != transfer_void_ptr:
            log.warning('Transfer mismatch')
            return
        if self._commands:
            self._finish(self._commands.pop(0), transfer)
        else:
            log.warning('Transfer callback when no commands')
        transfer.close()
        if self._handle is not None:
            self._issue()

    def _abort_all(self):
        commands, self._commands = self._commands, []
        status = self.stop_code if self.stop_code is not None else TransferStatus.CANCELLED
        for cbk_fn, setup_packet, _ in commands:
            try:
                response = usb_core.ControlTransferResponse(setup_packet, status, None)
                cbk_fn(response)
            except Exception:
                log.exception('in callback while aborting')

    def close(self):
        if self.stop_code is None:
            self.stop_code = 0
        handle, self._handle = self._handle, None
        if handle and self._transfer_pending:
            log.info('ControlTransferAsync.close cancel pending transfer, %d', len(self._commands))
            # callback function will be invoked later
            _lib.libusb_cancel_transfer(self._transfer_pending.transfer)
        else:
            log.info('ControlTransferAsync.close %d', len(self._commands))
            self._abort_all()

    def pend(self, cbk_fn, setup_packet: usb_core.SetupPacket, buffer=None):
        """Pend an asynchronous Control Transfer.

        :param cbk_fn: The function to call when the control transfer completes.
            A :class:`usb_core.ControlTransferResponse` is the sole argument.
        :param setup_packet:
        :param buffer: The buffer (if length > 0) for write transactions.
        :return: True if pending, False on error.
        """
        command = [cbk_fn, setup_packet, buffer]
        was_empty = not bool(self._commands)
        self._commands.append(command)
        if was_empty:
            return self._issue()
        return True

    def _issue(self):
        if not self._commands:
            return True
        if not self._handle:
            log.info('_issue but handle not valid')
            self._abort_all()
            return False
        if self.stop_code is not None:
            log.info('_issue but stop_code=%s', self.stop_code)
            self._abort_all()
            return False
        log.debug('preparing')
        _, setup_packet, buffer = self._commands[0]
        hdr = struct.pack('<BBHHH', setup_packet.request_type, setup_packet.request,
                          setup_packet.value, setup_packet.index, setup_packet.length)
        if buffer is not None:
            transfer = Transfer(hdr + buffer)
        else:
            transfer = Transfer(len(hdr) + setup_packet.length)
            transfer.buffer[:len(hdr)] = np.frombuffer(hdr, dtype=np.uint8)
        t = transfer.transfer[0]
        t.dev_handle = self._handle
        t.endpoint_id = 0
        t.endpoint_type = TransferType.CONTROL
        t.callback = self._transfer_callback_fn
        self._transfer_pending = transfer
        self._time_start = time.time()
        rv = _lib.libusb_submit_transfer(transfer.transfer)
        if 0 == rv:
            log.debug('libusb_submit_transfer [control]')
        else:
            log.warning('libusb_submit_transfer [control] => %d', rv)
            if t.status == 0:
                if rv == ReturnCodes.ERROR_NO_DEVICE:
                    log.info('control transfer but no device')
                    t.status = TransferStatus.NO_DEVICE
                else:
                    t.status = TransferStatus.ERROR
                if self.stop_code is None:
                    self.stop_code = DeviceEvent.COMMUNICATION_ERROR
            self._transfer_callback(transfer.addr)
            return False
        return True

    def _finish(self, command, transfer):
        buffer = None
        rc = transfer.transfer[0].status
        cbk_fn, setup_packet, _ = command
        pkt = usb_core.RequestType(value=setup_packet.request_type)
        duration = time.time() - self._time_start
        if rc == TransferStatus.NO_DEVICE:
            log.warning('device_removed')
            if self.stop_code is None:
                self.stop_code = DeviceEvent.COMMUNICATION_ERROR
        if pkt.direction == 'out':
            log.debug('ControlTransferAsync._finish rc=%d, duration=%.6f s', rc, duration)
        else:
            actual_length = transfer.transfer[0].actual_length
            log.debug('ControlTransferAsync._finish rc=%d, duration=%.6f s, length: %s, %s',
                      rc, duration, setup_packet.length, actual_length)
            buffer = bytes(transfer.buffer[8:(actual_length+8)])
        response = usb_core.ControlTransferResponse(setup_packet, rc, buffer)
        cbk_fn(response)


class EndpointIn:

    def __init__(self, handle, pipe_id, transfers, block_size, data_fn, process_fn, stop_fn):
        """Manage an in endpoint.

        :param handle: The device handle.
        :param pipe_id: The endpoint IN pipe identifier.
        :param transfers: The number of outstanding transfers to pend.
        :param block_size: The size of each transfer in bytes.
        :param data_fn: The function to call with the received endpoint IN data.
            After the last block, data_fn is called with None to indicate the
            last transfer.  The data_fn should normally return True, but can
            return False to stop the endpoint streaming.
        :param process_fn: The function() called after data_fn was called.
            This function can have more latency than data_fn.
        :param stop_fn: The function(event, message) called when this endpoint
            stops streaming data.
        """
        self._handle = handle
        self.pipe_id = pipe_id  # (endpoint_id & 0x7f) | 0x80
        self._config = {
            'transfer_count': transfers,
            'transfer_size_bytes': (block_size + 511 // 512)
        }
        self._data_fn = data_fn
        self._process_fn = process_fn
        self._stop_fn = stop_fn
        self.stop_code = None
        self.stop_message = ''

        self._state = self.ST_IDLE
        self._time_last = None
        self._transfers_free = []  # Transfer
        self._transfers_pending = []  # Transfer
        self.transfers_processed = 0

        self.transfer_count = 0
        self.byte_count_window = 0
        self.byte_count_total = 0
        self._transfer_callback_fn = libusb_transfer_cb_fn(self._transfer_callback)
        self._init()

    ST_IDLE = 0
    ST_RUNNING = 1
    ST_STOPPING = 2

    def __str__(self):
        return 'EndpointIn(0x%02x)' % (self.pipe_id, )

    def __len__(self):
        return len(self._transfers_pending)

    @property
    def is_busy(self):
        return 0 != len(self._transfers_pending)

    def _transfer_pending_pop(self, transfer_void_ptr):
        for idx in range(len(self._transfers_pending)):
            if transfer_void_ptr == self._transfers_pending[idx].addr:
                return self._transfers_pending.pop(idx)
        log.warning('%s _transfer_pending_pop not found', self)
        raise IOError('%s _transfer_pending_pop not found' % (self, ))

    def _transfer_done(self):
        if self.stop_code is None:
            log.warning('%s transfer_done by stop_code not set', self)
            self.stop_code = 0
        try:
            self._data_fn(None)
        except Exception:
            log.exception('_data_fn exception: stop streaming')
        try:
            self._stop_fn(self.stop_code, self.stop_message)
        except Exception:
            log.exception('_stop_fn exception')
        for transfer in self._transfers_free:
            transfer.close()
        self._transfers_free.clear()
        self._state = self.ST_IDLE
        log.info('%s transfer_done %d: %s', self, self.stop_code, self.stop_message)

    def _transfer_callback(self, transfer_void_ptr):
        transfer = self._transfer_pending_pop(transfer_void_ptr)
        self.transfer_count += 1
        t = transfer.transfer[0]

        try:
            if self._state == self.ST_RUNNING:
                if t.status == TransferStatus.COMPLETED:
                    self.byte_count_window += t.actual_length
                    self.byte_count_total += t.actual_length
                    self.transfers_processed += 1
                    buffer = transfer.buffer[:t.actual_length]
                    try:
                        rv = bool(self._data_fn(buffer))
                    except Exception:
                        log.exception('data_fn exception: stop streaming')
                        rv = True
                    if rv:
                        self._cancel(0, 'terminated by data_fn')
                elif t.status == TransferStatus.TIMED_OUT:
                    log.warning('%s: timed out', self)
                else:
                    msg = f'transfer callback with status {t.status}'
                    self._cancel(DeviceEvent.COMMUNICATION_ERROR, msg)
        finally:
            self._transfers_free.append(transfer)
        self._pend()
        if self._state == self.ST_STOPPING:
            if 0 == len(self._transfers_pending):
                self._transfer_done()
            else:
                log.debug('awaiting transfer completion')

    def _init(self):
        for i in range(self._config['transfer_count']):
            transfer = Transfer(self._config['transfer_size_bytes'])
            t = transfer.transfer[0]
            t.dev_handle = self._handle
            t.endpoint_id = self.pipe_id
            t.endpoint_type = TransferType.BULK
            t.callback = self._transfer_callback_fn
            self._transfers_free.append(transfer)

    def _pend(self):
        while self._state == self.ST_RUNNING and len(self._transfers_free):
            transfer = self._transfers_free.pop(0)
            transfer.transfer[0].status = TransferStatus.COMPLETED
            rv = _lib.libusb_submit_transfer(transfer.transfer)
            if rv:
                self._transfers_free.append(transfer)
                if rv in [ReturnCodes.ERROR_BUSY]:
                    log.info('libusb_submit_transfer busy')
                else:  # no device, not supported, or other error
                    msg = f'libusb_submit_transfer => {rv}'
                    self._cancel(DeviceEvent.COMMUNICATION_ERROR, msg)
                break  # give system time to recover
            else:
                self._transfers_pending.append(transfer)

    def _cancel(self, stop_code=None, stop_msg=None):
        if self._state != self.ST_RUNNING:
            return
        if self.stop_code is None:
            stop_code = 0 if stop_code is None else int(stop_code)
            stop_msg = '' if stop_msg is None else str(stop_msg)
            self.stop_code = stop_code
            self.stop_message = stop_msg
            lvl = logging.INFO if stop_code <= 0 else logging.ERROR
            log.log(lvl, 'endpoint halt %d: %s', stop_code, stop_msg)
        self._state = self.ST_STOPPING
        log.info('%s cancel %d : %d', self, self.stop_code, len(self._transfers_pending))
        for transfer in self._transfers_pending:
            _lib.libusb_cancel_transfer(transfer.transfer)
            # callbacks will be invoked later

    def process_signal(self):
        rv = False
        if self.transfer_count and self._state == self.ST_RUNNING:
            self.transfer_count = 0
            try:
                if callable(self._process_fn):
                    rv = bool(self._process_fn())
            except Exception:
                log.exception('_process_fn exception: stop streaming')
                rv = True  # force stop
            if rv:
                self._cancel(0, 'terminated by process_fn')
        return rv

    def start(self):
        log.info("%s start transfer size = %d bytes" % (self, self._config['transfer_size_bytes']))
        self.transfer_count = 0
        self.byte_count_window = 0
        self.byte_count_total = 0
        self.stop_code = None
        self.stop_message = None
        self._state = self.ST_RUNNING
        self._time_last = time.time()
        self._pend()
        time.sleep(0.0001)

    def stop(self):
        self._cancel(0, 'stop by method request')

    def status(self):
        """Get the endpoint status.

        :return: A dict mapping status name to a value..  The value is a dict
            containing 'value' and 'units'.
        """
        time_now = time.time()
        duration = time_now - self._time_last
        if duration < 0.01:
            throughput = 0.0
        else:
            throughput = self.byte_count_window / duration
        status = {
            'bytes': {'value': self.byte_count_total, 'units': 'bytes'},
            'transfers': {'value': self.transfer_count, 'units': 'transfers'},
            'duration': {'value': duration, 'units': 'seconds'},
            'throughput': {'value': throughput, 'units': 'Bps'},
        }
        self.byte_count_window = 0
        self._time_last = time_now
        return status


def may_raise_ioerror(rv, msg):
    if 0 != rv:
        s = msg + (' [%d]' % (rv, ))
        log.warning(s)
        raise IOError(s)
    else:
        log.debug('%s: success', msg.split(' ')[0])


class LibUsbDevice:
    """The LibUSB :class:`usb.api.Device` implementation"""

    def __init__(self, path):
        self._ctx = None
        self._path = path
        self._handle = c_void_p(None)
        self._endpoints = {}
        self._control_transfer = None  # type: ControlTransferAsync
        self._removed = False
        self._event_callback_fn = None

    def __str__(self):
        return f'Joulescope:{self.serial_number}'

    @property
    def serial_number(self):
        return self._path.split('/')[-1]

    def _open(self):
        log.info('open: start %s', self._path)
        self._ctx = _libusb_context_create()
        descriptor = _libusb_device_descriptor()
        devices = POINTER(POINTER(c_void_p))()
        vid, pid, serial_number = _path_split(self._path)
        sz = _lib.libusb_get_device_list(self._ctx, pointer(devices))
        try:
            for idx in range(sz):
                device = devices[idx]
                if self._handle:
                    pass
                elif _lib.libusb_get_device_descriptor(device, pointer(descriptor)):
                    pass
                elif vid == descriptor.idVendor and pid == descriptor.idProduct:
                    dh = c_void_p(None)
                    rv = _lib.libusb_open(device, dh)
                    if rv < 0:
                        log.info('Could not open device: %04x/%04x', vid, pid)
                    elif serial_number == _get_string_descriptor(dh, descriptor.iSerialNumber):
                        self._handle = dh
                        log.info('open: found device handle')
                        continue
                _lib.libusb_unref_device(device)
            if not self._handle:
                log.warning('open:failed')
                raise IOError('open:failed')
        finally:
            _lib.libusb_free_device_list(devices, 0)

    def open(self, event_callback_fn=None):
        # todo support event_callback_fn on errors
        self.close()
        try:
            self._open()
            log.info('open: configure device')
            rv = _lib.libusb_set_configuration(self._handle, 1)
            may_raise_ioerror(rv, 'libusb_set_configuration 1 failed')
            rv = _lib.libusb_claim_interface(self._handle, 0)
            may_raise_ioerror(rv, 'libusb_claim_interface 0 failed')
            rv = _lib.libusb_set_interface_alt_setting(self._handle, 0, 0)
            may_raise_ioerror(rv, 'libusb_set_interface_alt_setting 0,0 failed')
            self._control_transfer = ControlTransferAsync(self._handle)
            log.info('open: done')
        except IOError:
            log.exception('open failed: %s', self._path)
            self.close()
            raise
        except Exception as ex:
            log.exception('open failed: %s', self._path)
            self.close()
            raise IOError(ex)

    def _abort_endpoints(self):
        waiting = []
        if self._control_transfer is not None:
            self._control_transfer.close()
            waiting.append(self._control_transfer)
        for endpoint in self._endpoints.values():
            endpoint.stop()
            waiting.append(endpoint)
        time_start = time.time()
        while True:
            if all([not x.is_busy for x in waiting]):
                break
            dt = time.time() - time_start
            if dt > 0.25 + TRANSFER_TIMEOUT_MS / 1000:
                log.warning('Could not shut down gracefully')
                break
            timeval = TimeVal(tv_sec=0, tv_usec=25000)
            _lib.libusb_handle_events_timeout(self._ctx, byref(timeval))
        self._endpoints.clear()

    def close(self, status=None, message=None):
        log.info('close')
        self._abort_endpoints()
        if self._handle and not self._removed:
            handle, self._handle = self._handle, c_void_p(None)
            _lib.libusb_close(handle)
        if self._ctx:
            ctx, self._ctx = self._ctx, None
            _libusb_context_destroy(ctx)
        event_callback_fn, self._event_callback_fn = self._event_callback_fn, None
        if status is not None and callable(event_callback_fn):
            message = '' if message is None else str(message)
            try:
                event_callback_fn(status, message)
            except Exception:
                log.exception('while in _event_callback_fn')

    def _control_transfer_pend(self, cbk_fn, setup_packet, data):
        if self._control_transfer is None:
            rsp = usb_core.ControlTransferResponse(setup_packet, TransferStatus.NO_DEVICE, None)
            cbk_fn(rsp)
            return False
        return self._control_transfer.pend(cbk_fn, setup_packet, data)

    def control_transfer_out(self, cbk_fn, recipient, type_, request, value=0, index=0, data=None):
        if cbk_fn is None:
            run_until_done = RunUntilDone(1.0, 'control_transfer_out')
            self.control_transfer_out(run_until_done.cbk_fn, recipient, type_, request, value, index, data)
            while not run_until_done.is_done():
                self.process(0.01)
            return run_until_done.value_args0
        request_type = usb_core.RequestType(direction='out', type_=type_, recipient=recipient).u8
        length = 0 if data is None else len(data)
        setup_packet = usb_core.SetupPacket(request_type, request, value, index, length)
        return self._control_transfer_pend(cbk_fn, setup_packet, data)

    def control_transfer_in(self, cbk_fn, recipient, type_, request, value, index, length) -> ControlTransferResponse:
        if cbk_fn is None:
            run_until_done = RunUntilDone(1.0, 'control_transfer_in')
            self.control_transfer_in(run_until_done.cbk_fn, recipient, type_, request, value, index, length)
            while not run_until_done.is_done():
                self.process(0.01)
            return run_until_done.value_args0
        request_type = usb_core.RequestType(direction='in', type_=type_, recipient=recipient).u8
        setup_packet = usb_core.SetupPacket(request_type, request, value, index, length)
        return self._control_transfer_pend(cbk_fn, setup_packet, None)

    def read_stream_start(self, endpoint_id, transfers, block_size, data_fn, process_fn, stop_fn):
        pipe_id = (endpoint_id & 0x7f) | 0x80
        endpoint = self._endpoints.pop(pipe_id, None)
        if endpoint is not None:
            log.warning('repeated start')
            endpoint.stop()
        endpoint = EndpointIn(self._handle, pipe_id, transfers,
                              block_size, data_fn, process_fn, stop_fn)
        self._endpoints[endpoint.pipe_id] = endpoint
        endpoint.start()

    def read_stream_stop(self, endpoint_id):
        log.info('read_stream_stop %d', endpoint_id)
        pipe_id = (endpoint_id & 0x7f) | 0x80
        endpoint = self._endpoints.pop(pipe_id, None)
        if endpoint is not None:
            endpoint.stop()

    def status(self):
        e = {}
        s = {'endpoints': e}
        for endpoint in self._endpoints.values():
            e[endpoint.pipe_id] = endpoint.status()
        return s

    def signal(self):
        pass  # todo, currently delays in process for up to 25 ms waiting for libusb_handle_events_timeout

    def process(self, timeout=None):
        if self._ctx and not self._removed:
            timeval = TimeVal(tv_sec=0, tv_usec=25000)
            _lib.libusb_handle_events_timeout(self._ctx, byref(timeval))
            endpoints_stop = []
            for endpoint in self._endpoints.values():
                endpoint.process_signal()
                if endpoint._state == endpoint.ST_IDLE:
                    endpoints_stop.append(endpoint.pipe_id)
            for pipe_id in endpoints_stop:
                self._endpoints.pop(pipe_id, None)
            if self._control_transfer.stop_code:
                msg = f'Control endpoint failed {self._control_transfer.stop_code}'
                self.close(self._control_transfer.stop_code, msg)
        else:
            time.sleep(0.025)


class DeviceNotify:

    def __init__(self, cbk):
        self._cbk = cbk
        self._window_thread = None
        self._do_quit = True
        if 0 == _lib.libusb_has_capability(Capability.HAS_HOTPLUG):
            log.warning('libusb does not support hotplug')
            return
        self.open()

    def _hotplug_cbk(self, ctx, device, ev, user_data):
        inserted = bool(ev & HotplugEvent.DEVICE_ARRIVED)
        removed = bool(ev & HotplugEvent.DEVICE_LEFT)
        log.info('hotplug: inserted=%s, removed=%s', inserted, removed)
        self._cbk(inserted, 0)
        return 0

    def _run(self):
        if 0 == _lib.libusb_has_capability(Capability.HAS_HOTPLUG):
            log.warning('libusb does not support hotplug')
            # todo revert to polling method?
            return
        log.debug('_run_window start')
        timeval = TimeVal(tv_sec=0, tv_usec=100000)
        timeval_ptr = pointer(timeval)
        handle = c_int()
        cbk_fn = _libusb_hotplug_callback_fn(self._hotplug_cbk)
        cbk_user_data = c_void_p()
        with _libusb_context() as ctx:
            rv = _lib.libusb_hotplug_register_callback(
                ctx,
                HotplugEvent.DEVICE_ARRIVED | HotplugEvent.DEVICE_LEFT,
                0, # flags
                HOTPLUG_MATCH_ANY,  # vid
                HOTPLUG_MATCH_ANY,  # pid
                HOTPLUG_MATCH_ANY,  # device class
                cbk_fn,
                cbk_user_data,
                byref(handle))
            if rv:
                raise IOError('could not register hotplug')
            while not self._do_quit:
                _lib.libusb_handle_events_timeout(ctx, timeval_ptr)
            _lib.libusb_hotplug_deregister_callback(ctx, handle)

    def open(self):
        self.close()
        self._do_quit = False
        log.info('open')
        self._window_thread = threading.Thread(name='device_notify', target=self._run)
        self._window_thread.start()

    def close(self):
        if self._window_thread:
            log.info('close')
            self._do_quit = True
            self._window_thread.join()
            self._window_thread = None


def scan(name: str) -> List[LibUsbDevice]:
    """Scan for attached devices.

    :param name: The case-insensitive name of the device to scan.
    :return: The list of discovered :class:`WinUsbDevice` instances.
    """
    with _libusb_context() as ctx:
        paths = []
        infos = INFO[name.lower()]
        descriptor = _libusb_device_descriptor()
        devices = POINTER(POINTER(c_void_p))()
        sz = _lib.libusb_get_device_list(ctx, pointer(devices))
        try:
            for idx in range(sz):
                device = devices[idx]
                if _lib.libusb_get_device_descriptor(device, pointer(descriptor)):
                    raise RuntimeError('descriptor')
                for info in infos:
                    vid = info['vendor_id']
                    pid = info['product_id']
                    if vid == descriptor.idVendor and pid == descriptor.idProduct:
                        dh = c_void_p(None)
                        rv = _lib.libusb_open(device, pointer(dh))
                        if rv < 0 or not dh:
                            log.info('Could not open device: %04x/%04x', vid, pid)
                            continue
                        try:
                            serial_number = _get_string_descriptor(dh, descriptor.iSerialNumber)
                        finally:
                            _lib.libusb_close(dh)
                        path = '%04x/%04x/%s' % (vid, pid, serial_number)
                        paths.append(path)
        finally:
            _lib.libusb_free_device_list(devices, 1)

        if not len(paths):
            log.info('scan found no devices')
            return []
        log.info('scan found %s' % paths)
        devices = [LibUsbDevice(p) for p in paths]
        return devices


def _on_device_notify(inserted, info):
    print(f'_on_device_notify({inserted}, {info})')


if __name__ == '__main__':
    logging.basicConfig(level=logging.DEBUG)
    devices = scan('Joulescope')
    print('\n'.join([str(d) for d in devices]))
    if len(devices):
        d = devices[0]
        d.open()
        rv = d.control_transfer_in(None, 'device', 'vendor', request=4,
            value=0, index=0, length=128)
        print(rv)
        d.close()



# Copyright 2018-2022 Jetperch LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Stream Buffer implementation for v1 backend.
"""

from .sample_buffer import SampleBuffer
from .stats import compute_stats
import numpy as np
import logging


# statistics format for numpy structured data type
# https://docs.scipy.org/doc/numpy/user/basics.rec.html
_STATS_FIELDS = 6  # current, voltage, power, current_range, current_lsb, voltage_lsb
NP_STATS_FORMAT = ['u8', 'f8', 'f8', 'f8', 'f8']
NP_STATS_NAMES = ['length', 'mean', 'variance', 'min', 'max']
STATS_FIELD_NAMES = ['current', 'voltage', 'power', 'current_range', 'current_lsb', 'voltage_lsb']
STATS_DTYPE = np.dtype({'names': NP_STATS_NAMES, 'formats': NP_STATS_FORMAT})
STATS_FIELD_COUNT = _STATS_FIELDS
FIELDS = {
    'current':       [(1, 0), 'A'],
    'voltage':       [(2, 0), 'V'],
    'power':         [(3, 0), 'W'],
    'current_range': [(4, 0), ''],
    'current_lsb':   [(5, 0), ''],
    'voltage_lsb':   [(5, 1), ''],
}

class StreamBuffer:
    """Efficient real-time Joulescope data buffering.

    :param duration: The total length of the buffering in seconds.
    """

    def __init__(self, duration, frequency, device, output_frequency=None):
        # current, voltage, power, current_range, gpi0, gpi1
        self._input_sampling_frequency = frequency
        if output_frequency is not None:
            self._sampling_frequency = output_frequency
        else:
            self._sampling_frequency = frequency
        self._buffer_duration = duration  #: total buffer duration in seconds
        self._device = device
        self._log = logging.getLogger(__name__)
        self._length = 0
        self.buffers = {}
        self._duration_max = 0
        self._contiguous_duration_max = 0
        self._sample_id_start = None
        self._callback = None
        self._update()

    def _update(self):
        self._length = int(self._buffer_duration * self._sampling_frequency)
        self.buffers.clear()
        self._sample_id_start = None
        self.buffers = {
            # (field_id, index): SampleBuffer
            (1, 0): SampleBuffer(self._length, dtype=np.float32, name='current'),
            (2, 0): SampleBuffer(self._length, dtype=np.float32, name='voltage'),
            (3, 0): SampleBuffer(self._length, dtype=np.float32, name='power'),
            (4, 0): SampleBuffer(self._length, dtype=np.uint8, name='current_range'),  # (converted u4->u8 by pyjoulescope_driver binding)
            (5, 0): SampleBuffer(self._length, dtype='u1', name='gpi0'),
            (5, 1): SampleBuffer(self._length, dtype='u1', name='gpi1'),
        }

    def __len__(self):
        return self._length

    def __str__(self):
        return f'StreamBuffer(length={self._length}, reductions=[])'

    @property
    def voltage_range(self):
        return 0

    @property
    def has_raw(self):
        return False

    @property
    def sample_id_range(self):
        """Get the range of sample ids currently available in the buffer.

        :return: Tuple of sample_id start, sample_id end.
            Start and stop follow normal python indexing:
            start is inclusive, end is exclusive
        """
        r = [b.range for b in self.buffers.values() if b.active]
        if len(r) == 0:
            return None, None
        start = max([b[0] for b in r])
        end = min([b[1] for b in r])
        if end < start:
            end = start
        return start, end

    @property
    def duration_max(self):
        return self._duration_max

    @duration_max.setter
    def duration_max(self, value):
        self._duration_max = value  # stop streaming when reach this sample

    @property
    def is_duration_max(self):
        if self._duration_max is None:
            return False
        e1, e2 = self.sample_id_range
        if e1 is None or self._sample_id_start is None:
            return False
        duration = (e2 - self._sample_id_start) / self._sampling_frequency
        if duration < self._duration_max:
            return False
        return True

    @property
    def contiguous_duration_max(self):
        return self._contiguous_duration_max

    @contiguous_duration_max.setter
    def contiguous_duration_max(self, value):
        self._contiguous_duration_max = value

    @property
    def is_contiguous_duration_max(self):
        if self._contiguous_duration_max is None:
            return False
        # todo: reset on sample drops
        e1, e2 = self.sample_id_range
        if e1 is None or self._sample_id_start is None:
            return False
        duration = (e2 - self._sample_id_start) / self._sampling_frequency
        if duration < self._contiguous_duration_max:
            return False
        return True

    @property
    def callback(self):
        return self._callback

    @callback.setter
    def callback(self, value):
        self._callback = value

    @property
    def input_sampling_frequency(self):
        return self._input_sampling_frequency

    @property
    def output_sampling_frequency(self):
        return self._sampling_frequency

    @output_sampling_frequency.setter
    def output_sampling_frequency(self, value):
        self._sampling_frequency = value
        self._update()

    @property
    def buffer_duration(self):
        return self._buffer_duration

    @buffer_duration.setter
    def buffer_duration(self, value):
        self._buffer_duration = value
        self._update()

    @property
    def limits_time(self):
        return 0.0, len(self) / self._sampling_frequency

    @property
    def limits_samples(self):
        _, s_max = self.sample_id_range
        return (s_max - len(self)), s_max

    def time_to_sample_id(self, t):
        idx_start, idx_end = self.limits_samples
        t_start, t_end = self.limits_time
        return int(round((float(t) - t_start) / (t_end - t_start) * (idx_end - idx_start) + idx_start))

    def sample_id_to_time(self, s):
        idx_start, idx_end = self.limits_samples
        t_start, t_end = self.limits_time
        return (int(s) - idx_start) / (idx_end - idx_start) * (t_end - t_start) + t_start

    def status(self):
        return {
            'device_sample_id': {'value': 0, 'units': 'samples'},
            # 'sample_id': {'value': self._sample_id_max, 'units': 'samples'},
            'sample_missing_count': {'value': 0, 'units': 'samples'},
            'skip_count': {'value': 0, 'units': ''},
            'sample_sync_count': {'value': 0, 'units': 'samples'},
            'contiguous_count': {'value': 0, 'units': 'samples'},
        }

    def calibration_set(self, current_offset, current_gain, voltage_offset, voltage_gain):
        pass

    def reset(self):
        for b in self.buffers.values():
            b.clear()
        self._sample_id_start = None

    def insert(self, topic, value):
        try:
            idx = (value['field_id'], value['index'])
            b = self.buffers[idx]
            sample_rate = value.get('sample_rate')
            decimate_factor = value.get('decimate_factor')
            local_decimate = 1
            if decimate_factor == 1 and self._sampling_frequency != sample_rate:
                local_decimate = sample_rate // self._sampling_frequency
            b.add(value['sample_id'], value['data'],
                  sample_rate=sample_rate,
                  decimate_factor=decimate_factor,
                  local_decimate_factor=local_decimate)
            if self._sample_id_start is None:
                e1, e2 = self.sample_id_range
                if e1 is not None and e1 != e2:
                    self._sample_id_start = e1
            # print(f'{topic} {value["field_id"]}.{value["index"]} {value["sample_id"]} {len(value["data"])} {sample_rate} {decimate_factor} {local_decimate}')
        except KeyError:
            pass  # buffer does not exist, skip
        return 0

    def statistics_get(self, start, stop, out=None):
        """Get exact statistics over the specified range.

        :param start: The starting sample_id (inclusive).
        :param stop: The ending sample_id (exclusive).
        :param out: The optional output np.ndarray(STATS_FIELD_COUNT, dtype=DTYPE).
            None (default) creates and outputs a new record.
        :return: The tuple of (np.ndarray(STATS_FIELD_COUNT, dtype=DTYPE), [start, stop]).
            The values of start and stop will be constrained to the
            available range.
        """
        self_start, self_stop = self.sample_id_range
        #self._log.debug(f'statistics_get({start}, {stop}) in ({self_start}, {self_stop})')
        start = max(start, self_start)
        stop = min(stop, self_stop)
        if out is None:
            out = np.zeros(_STATS_FIELDS, dtype=STATS_DTYPE)
        if stop >= self_start and start < self_stop:
            out[:]['length'] = stop - start
            for i, b in enumerate(self.buffers.values()):
                if not b.active:
                    out[i]['length'] = 0
                    out[i]['mean'] = np.nan
                    out[i]['variance'] = np.nan
                    out[i]['min'] = np.nan
                    out[i]['max'] = np.nan
                    continue
                d = b.get_range(start, stop)
                compute_stats(d, out[i])
        else:
            out[:]['length'] = 0
            out[:]['mean'] = np.nan
            out[:]['variance'] = np.nan
            out[:]['min'] = np.nan
            out[:]['max'] = np.nan
        return out, (start, stop)

    def data_get(self, start, stop, increment=None, out=None):
        """Get the samples with statistics.

        :param start: The starting sample id (inclusive).
        :param stop: The ending sample id (exclusive).
        :param increment: The number of raw samples.
        :param out: The optional output np.ndarray((N, STATS_FIELD_COUNT), dtype=DTYPE) to populate with
            the result.  None (default) will construct and return a new array.
        :return: The np.ndarray((N, STATS_FIELD_COUNT), dtype=DTYPE) data.

        This method provides fast access to statistics data, primarily for
        graphical display.  Applications should prefer using
        :meth:`samples_get` which provides metadata along with the samples.
        """
        self_start, self_stop = self.sample_id_range
        # self._log.debug(f'data_get({start}, {stop}, {increment}) in ({self_start}, {self_stop})')
        expected_length = (stop - start) // increment
        n_total = (stop - start) // increment
        if out is not None:
            if len(out) < expected_length:
                raise ValueError('out too small')
        else:
            out = np.zeros((expected_length, _STATS_FIELDS), dtype=STATS_DTYPE)

        for n in range(n_total):
            k_start = start + n * increment
            k_end = k_start + increment
            if k_start < self_start or k_end > self_stop:
                out[n, :]['length'] = 0
                out[n, :]['mean'] = np.nan
                out[n, :]['variance'] = np.nan
                out[n, :]['min'] = np.nan
                out[n, :]['max'] = np.nan
                continue
            for i, b in enumerate(self.buffers.values()):
                if not b.active:
                    out[n, i]['length'] = increment
                    out[n, i]['mean'] = np.nan
                    out[n, i]['variance'] = np.nan
                    out[n, i]['min'] = np.nan
                    out[n, i]['max'] = np.nan
                try:
                    d = b.get_range(k_start, k_end)
                    compute_stats(d, out[n, i])
                except KeyError:
                    pass
        return out

    def samples_get(self, start, stop, fields=None):
        """Get exact sample data without any skips or reductions.

        :param start: The starting sample id (inclusive).
        :param stop: The ending sample id (exclusive).
        :param fields: The single field or list of field names to return.
            None (default) is equivalent to
            ['current', 'voltage', 'power', 'current_range', 'current_lsb', 'voltage_lsb'].
            The available fields are:

            * raw: The raw u16 data from Joulescope.
              Equivalent to self.raw_get(start, stop)
            * raw_current: The raw 14-bit current data in LSBs.
            * raw_voltage: The raw 14-bit voltage data in LSBs.
            * current: The calibrated float32 current data array in amperes.
            * voltage: The calibrated float32 voltage data array in volts.
            * current_voltage: The calibrated float32 Nx2 array of current, voltage.
            * power: The calibrated float32 power data array in watts.
            * bits: The (voltage_lsb << 5) | (current_lsb << 4) | current_range
            * current_range: The current range. 0 = 10A, 6 = 18 uA, 7=off.
            * current_lsb: The current LSB, which can be assign to a general purpose input.
            * voltage_lsb: The voltage LSB, which can be assign to a general purpose input.

        :return: The dict containing top-level 'time' and 'signals' keys.
            The 'time' value is a dict contain the timing metadata for
            these samples.  The 'signals' value is a dict with one
            key for each field in fields.  Each field value is also
            a dict with entries 'value' and 'units'.
            However, if single field string is provided to fields, then just
            return that field's value.
        """
        if fields is None:
            fields = ['current', 'voltage', 'power', 'current_range', 'current_lsb', 'voltage_lsb']
        self_start, self_stop = self.sample_id_range
        #self._log.debug(f'samples_get({start}, {stop}) in ({self_start}, {self_stop})')
        start = max(start, self_start)
        stop = min(stop, self_stop)
        t1 = self.sample_id_to_time(start)
        t2 = self.sample_id_to_time(stop)
        result = {
            'time': {
                'range': {'value': [t1, t2], 'units': 's'},
                'delta': {'value': t2 - t1, 'units': 's'},
                'sample_id_range': {'value': [start, stop], 'units': 'samples'},
                'samples': {'value': stop - start, 'units': 'samples'},
                'input_sampling_frequency': {'value': self.input_sampling_frequency, 'units': 'Hz'},
                'output_sampling_frequency': {'value': self.output_sampling_frequency, 'units': 'Hz'},
                'sampling_frequency': {'value': self.output_sampling_frequency, 'units': 'Hz'},
            },
            'signals': {},
        }
        for field in fields:
            try:
                idx, units = FIELDS[field]
                b = self.buffers.get(idx)
                if b.active:
                    out = b.get_range(start, stop)
                    result['signals'][field] = {'value': out, 'units': units}
                else:
                    d = np.empty(stop - start, dtype=np.float32)
                    d[:] = np.nan
                    result['signals'][field] = {'value': out, 'units': units}
            except KeyError:
                pass  # cannot include this signal
        return result



# Copyright 2018 Jetperch LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from joulescope import datafile
import io
import struct
import binascii
import logging
log = logging.getLogger(__name__)

DeviceInterfaceGUID = '{76ab0534-eeda-48b5-b0c9-617fc6ce8296}'
VENDOR_ID = 0x1FC9
PRODUCT_ID = 0xFC94
HOST_API_VERSION = 1
INFO_REQUEST_LENGTH = 256
CHUNK_SIZE = 0x1000


def _ioerror_on_bad_result(rv):
    if 0 != rv.result:
        raise IOError('usb.Device %s' % (rv,))


class UsbdRequest:
    LOOPBACK_WVALUE = 1
    LOOPBACK_BUFFER = 2
    INFO = 3
    """Get the detailed device information"""

    WRITE_START = 4
    WRITE_CHUNK = 5
    WRITE_FINALIZE = 6
    READ_CHUNK = 7
    GO = 8
    ERASE = 9


class Segment:
    PERSONALITY = 1
    FIRMWARE = 2
    STORAGE1 = 3
    CALIBRATION_ACTIVE = 4
    CALIBRATION_FACTORY = 5


SEGMENTS = {
    'calibration_factory': Segment.CALIBRATION_FACTORY,
    'firmware': Segment.FIRMWARE,
    'storage1': Segment.STORAGE1,
    'personality': Segment.PERSONALITY,
    'calibration_active': Segment.CALIBRATION_ACTIVE,
}
for segments_value in list(SEGMENTS.values()):
    SEGMENTS[segments_value] = segments_value


class Bootloader:
    """The bootloader API and implementation for use by applications.

    :param usb_device: The backend USB :class:`usb.device` instance.
    """
    def __init__(self, usb_device):
        self._usb = usb_device
        self._is_open = False

    def __str__(self):
        return self._usb.serial_number

    @property
    def usb_device(self):
        """Get the USB backend device implementation.

        This method should only be used for unit and system tests.  Production
        code should *NEVER* access the underlying USB device directly.
        """
        return self._usb

    def open(self):
        """Open the device for use

        :raise IOError: on failure.
        """
        if self._is_open:
            self.close()
        self._is_open = True
        self._usb.open()

    def close(self):
        """Close the device and release resources"""
        self._is_open = False
        try:
            self._usb.close()
        except Exception:
            log.exception('USB close failed')

    def comm_test(self):
        rv = self._usb.control_transfer_out(
            None, 'device', 'vendor',
            request=UsbdRequest.LOOPBACK_WVALUE,
            value=0x1234, index=0)
        _ioerror_on_bad_result(rv)
        rv = self._usb.control_transfer_in(
            None, 'device', 'vendor',
            request=UsbdRequest.LOOPBACK_WVALUE,
            value=0, index=0, length=2)
        _ioerror_on_bad_result(rv)
        value = struct.unpack('<H', bytes(rv.data))[0]
        print(value)

    def info_get(self):
        rv = self._usb.control_transfer_in(
            None, 'device', 'vendor',
            request=UsbdRequest.INFO,
            value=0, index=0, length=INFO_REQUEST_LENGTH)
        _ioerror_on_bad_result(rv)
        values = struct.unpack('<IIIII', bytes(rv.data))
        names = ['bootloader_version', 'hardware_id', 'hardware_version', 'firmware_version', 'status']
        data_str = ['%s=0x%08x' % (name, value) for name, value in zip(names, values)]
        log.info('\n    '.join(['info_get:'] + data_str))
        return dict(zip(names, values))

    def chunk_read(self, segment, chunk):
        """Read a data chunk segment from the device.

        :param segment: The segment number.
        :param chunk: The chunk number within the segment.
        :return: The data read from the device.
        :raise IOError: on any error, including reading segments
            that do not have read permissions.
        """
        rv = self._usb.control_transfer_in(
            None, 'device', 'vendor',
            request=UsbdRequest.READ_CHUNK,
            value=segment, index=chunk, length=CHUNK_SIZE)
        _ioerror_on_bad_result(rv)
        return bytes(rv.data)

    def program(self, segment, data, metadata=None, progress_cbk=None):
        """Program a segment with data.

        :param segment: The segment to program in flash, either integer
            identifier or name from SEGMENTS.
        :param data: The raw data bytes for the flash.
        :param metadata: The metadata for firmware updates.
        :param progress_cbk:  The optional Callable[[float], None] which is called
            with the progress fraction from 0.0 to 1.0
        :return: 0 on success or error code.

        This function erases the existing application, even on failure.
        """
        log.info('%s: program', self)
        data = datafile.filename_or_bytes(data)
        segment = SEGMENTS[segment]
        if metadata is None:
            metadata = {
                'encryption': 0,
                'header': bytes([0] * 24),
                'mac': bytes([0] * 16),
                'signature': bytes([0] * 64),
            }
        if progress_cbk is None:
            progress_cbk = lambda x: None

        progress_cbk(0.000)
        metadata['size'] = len(data)
        msg = struct.pack('<II', metadata['size'], metadata['encryption'])
        msg = msg + metadata['header'] + metadata['mac'] + metadata['signature']
        log.info('write start: segment=%d, length=%d', segment, len(data))
        rv = self._usb.control_transfer_out(
            None, 'device', 'vendor',
            request=UsbdRequest.WRITE_START,
            value=segment, index=0, data=msg)
        _ioerror_on_bad_result(rv)

        total_size = len(data)
        chunk = 0
        while len(data):
            fraction_done = chunk * CHUNK_SIZE / total_size
            chunk_data = data[:CHUNK_SIZE]
            progress_cbk(fraction_done)
            log.info('write chunk %d, length=%d | %.1f%%', chunk, len(chunk_data), fraction_done * 100)
            rv = self._usb.control_transfer_out(
                None, 'device', 'vendor',
                request=UsbdRequest.WRITE_CHUNK,
                value=segment, index=chunk, data=chunk_data)
            _ioerror_on_bad_result(rv)
            data = data[CHUNK_SIZE:]
            chunk += 1
        log.info('write finalize')
        rv = self._usb.control_transfer_in(
            None, 'device', 'vendor',
            request=UsbdRequest.WRITE_FINALIZE,
            value=segment, index=0, length=1)
        _ioerror_on_bad_result(rv)
        log.info('write status=%d', rv.data[0])
        progress_cbk(1.0)
        return rv.data[0]

    def go(self):
        """Command the bootloader to run the application.

        :raise IOError: on failure.

        This method will close() the bootloader regardless of success.
        """
        log.info('%s: go', self)
        try:
            rv = self._usb.control_transfer_out(
                None, 'device', 'vendor',
                request=UsbdRequest.GO,
                value=0, index=0)
            _ioerror_on_bad_result(rv)
        finally:
            self.close()

    def firmware_program(self, filename, progress_cbk=None):
        """Program or update the application firmware.

        :param filename: The filename containing the firmware, which must be
            in Joulescope firmware format and correctly signed.
        :param progress_cbk:  The optional Callable[[float], None] which is called
            with the progress fraction from 0.0 to 1.0
        :return: 0 on success or error code.
        """
        log.info('%s: firmware_program', self)
        data = datafile.filename_or_bytes(filename)
        if len(data):
            fh = io.BytesIO(data)
            dr = datafile.DataFileReader(fh)
            # todo: check distribution signature
            tag, hdr_value = next(dr)
            if tag != datafile.TAG_HEADER:
                raise ValueError('incorrect format: expected header, received %r' % tag)
            tag, data = next(dr)
            if tag != datafile.TAG_DATA_BINARY:
                raise ValueError('incorrect format: expected data, received %r' % tag)
            tag, enc = next(dr)
            if tag != datafile.TAG_ENCRYPTION:
                raise ValueError('incorrect format: expected encryption, received %r' % tag)
            metadata = {
                'encryption': 1,
                'header': hdr_value[:24],
                'mac': enc[:16],
                'signature': enc[16:],
            }
            log.info('header    = %r', binascii.hexlify(metadata['header']))
            log.info('mac       = %r', binascii.hexlify(metadata['mac']))
            log.info('signature = %r', binascii.hexlify(metadata['signature']))
        else:
            metadata = None
        return self.program(Segment.FIRMWARE, data, metadata, progress_cbk)

    def calibration_program(self, filename, is_factory=False):
        """Program the calibration.

        :param filename: The filename containing the Joulescope calibration
            record for this device.
        :param is_factory: True to program factory calibration.  False to
            program the active calibration.  The device only allows
            factory calibration to be programmed once.  Attempting to
            overwrite factory calibration will result in an error.
        :return: 0 on success or error code.
        """
        log.info('%s: calibration_program', self)
        data = datafile.filename_or_bytes(filename)
        segment = Segment.CALIBRATION_FACTORY if bool(is_factory) else Segment.CALIBRATION_ACTIVE
        return self.program(segment, data)
        
    def _erase(self, magic, sector_start=None, sector_end=None):
        """Permanently erase flash on Joulescope.
        
        :param magic: The 32-byte erase key.
        :param sector_start: The starting sector to erase.
        :param sector_end: The ending sector to erase.
        
        WARNING: This operation renders Joulescope useless and should never be
        used other than during manufacturing.
        """
        if sector_end < sector_start:  # swap
            sector_start, sector_end = sector_end, sector_start
        sector_start = 0 if sector_start is None else int(sector_start)
        sector_end = 15 if sector_end is None else int(sector_end)
        if not 0 <= sector_start <= 15:
            raise ValueError('erase sector_start out of range: %s', sector_start)
        if not 0 <= sector_end <= 15:
            raise ValueError('erase sector_end out of range: %s', sector_end)
        log.info('%s ERASE (%d, %d)', self, sector_start, sector_end)
        # arm
        data = struct.pack('<III', 1, 0, 0)
        rv = self._usb.control_transfer_out(
            None, 'device', 'vendor',
            request=UsbdRequest.ERASE,
            value=0, index=0, data=magic + data)        
        _ioerror_on_bad_result(rv)
        
        # and do it... hope you meant it!
        data = struct.pack('<III', 2, sector_start, sector_end)
        rv = self._usb.control_transfer_out(
            None, 'device', 'vendor',
            request=UsbdRequest.ERASE,
            value=0, index=0, data=magic + data)        
        _ioerror_on_bad_result(rv)        
        return 0

    def erase(self, magic, operation=None):
        """Permanently erase flash on a Joulescope.  !!! WARNING: factory use only!!!

        :param magic: The 32-byte erase key.
        :param operation: The erase operation which is one of:
            * all: (default) Delete the everything: The application,
              bootloader, calibration, and personalization.
            * calibration: Delete all calibration information and storage.
        """
        sector_start, sector_end = {
            'all': (0, 15),
            'calibration': (13, 15),
        }.get(operation, (0, 15))
        return self._erase(magic, sector_start, sector_end)



# Copyright 2022-2023 Jetperch LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from joulescope.parameters_v1 import PARAMETERS, PARAMETERS_DICT, name_to_value, value_to_name
from .stream_buffer import StreamBuffer
from joulescope.view import View
import copy
import logging
import numpy as np
import queue
import time


class Device:

    def __init__(self, driver, device_path):
        self.config = None
        self._driver = driver
        while device_path.endswith('/'):
            device_path = device_path[:-1]
        self._log = logging.getLogger(__name__ + '.' + device_path.replace('/', '.'))
        self._path = device_path
        self.is_open = False
        self._stream_cbk_objs = []
        self._stream_cbk_objs_add = []
        self._stop_fn = None
        self._input_sampling_frequency = 0
        self._output_sampling_frequency = 0
        self._statistics_callbacks = []
        self._statistics_offsets = []
        self._is_streaming = False
        self._stream_topics = []
        self._buffer_duration = 30
        self.stream_buffer = None
        self._on_stats_cbk = self._on_stats  # hold reference for unsub
        self._on_stream_cbk = self._on_stream  # hold reference for unsub
        self._parameters = {}
        self._parameter_set_queue = []
        for p in PARAMETERS:
            if p.default is not None:
                self._parameters[p.name] = name_to_value(p.name, p.default)

    def __str__(self):
        _, model, serial_number = self._path.split('/')
        return f'{model.upper()}-{serial_number}'

    @property
    def device_path(self):
        return self._path

    @property
    def usb_device(self):
        return self._path

    @property
    def input_sampling_frequency(self):
        """The original input sampling frequency."""
        return self._input_sampling_frequency

    @property
    def output_sampling_frequency(self):
        """The output sampling frequency."""
        return self._output_sampling_frequency

    @output_sampling_frequency.setter
    def output_sampling_frequency(self, value):
        self._output_sampling_frequency = value
        if self.stream_buffer is not None:
            self.stream_buffer.output_sampling_frequency = self._output_sampling_frequency

    @property
    def sampling_frequency(self):
        """The output sampling frequency."""
        return self.output_sampling_frequency

    @property
    def buffer_duration(self):
        """The stream buffer duration."""
        return self._buffer_duration

    @buffer_duration.setter
    def buffer_duration(self, value):
        self._buffer_duration = value
        if self.stream_buffer is not None:
            self.stream_buffer.buffer_duration = self._buffer_duration

    @property
    def statistics_callback(self):
        """Get the registered statistics callback."""
        cbks = self._statistics_callbacks
        if len(cbks):
            return cbks[0]
        else:
            return None

    @statistics_callback.setter
    def statistics_callback(self, cbk):
        """Set the statistics callback.

        :param cbk: The callable(data) where data is a statistics data
            structure.  See the `statistics documentation <statistics.html>`_
            for details on the data format.
            This function will be called from the USB processing thread.
            Any calls back into self MUST BE resynchronized.
        """
        for unregister_cbk in list(self._statistics_callbacks):
            self.statistics_callback_unregister(unregister_cbk)
        self.statistics_callback_register(cbk)

    def statistics_callback_register(self, cbk, source=None):
        """Register a statistics callback.

        :param cbk: The callable(data) where data is a statistics data
            structure.  See the `statistics documentation <statistics.html>`_
            for details on the data format.
            This function will be called from the USB processing thread.
            Any calls back into self MUST BE resynchronized.
        :param source: The statistics source where the computation is performed.
            Ignored, always use sensor-side statistics for the JS220.

        WARNING: calling :meth:`statistics_callback` after calling this method
        may result in unusual behavior.  Do not mix these API calls.
        """
        if cbk is None:
            return
        if not callable(cbk):
            self._log.warning('Requested callback is not callable')
            return
        if not len(self._statistics_callbacks):
            self._statistics_start()
        self._statistics_callbacks.append(cbk)

    def statistics_callback_unregister(self, cbk, source=None):
        """Unregister a statistics callback.

        :param cbk: The callback previously provided to
            :meth:`statistics_callback_register`.
        :param source: The callback source.
        """
        try:
            self._statistics_callbacks.remove(cbk)
        except ValueError:
            self._log.warning('statistics_callback_unregister but callback not registered.')
        if not len(self._statistics_callbacks):
            self._statistics_stop()

    def _on_stats(self, topic, value):
        period = 1 / 2e6
        s_start, s_stop = [x * period for x in value['time']['samples']['value']]

        if not len(self._statistics_offsets):
            duration = s_start
            charge = value['accumulators']['charge']['value']
            energy = value['accumulators']['energy']['value']
            offsets = [duration, charge, energy]
            self._statistics_offsets = [duration, charge, energy]
        duration, charge, energy = self._statistics_offsets
        value['time']['range'] = {
            'value': [s_start - duration, s_stop - duration],
            'units': 's'
        }
        value['time']['delta'] = {'value': s_stop - s_start, 'units': 's'}
        value['accumulators']['charge']['value'] -= charge
        value['accumulators']['energy']['value'] -= energy
        for k in value['signals'].values():
            k['µ'] = k['avg']
            k['σ2'] = {'value': k['std']['value'] ** 2, 'units': k['std']['units']}
            if 'integral' in k:
                k['∫'] = k['integral']
        for cbk in self._statistics_callbacks:
            cbk(value)

    def _statistics_start(self):
        if self.is_open:
            self.publish('s/stats/ctrl', 1)
            self.subscribe('s/stats/value', 'pub', self._on_stats_cbk)

    def _statistics_stop(self):
        if self.is_open:
            self.unsubscribe('s/stats/value', self._on_stats_cbk)
            self.publish('s/stats/ctrl', 0)

    def statistics_accumulators_clear(self):
        """Clear the charge and energy accumulators."""
        self._statistics_offsets.clear()

    def view_factory(self):
        """Construct a new View into the device's data.

        :return: A View-compatible instance.
        """
        if self.stream_buffer is None:
            raise RuntimeError('view_factory, but no stream buffer')
        view = View(self.stream_buffer, self.calibration)
        view.on_close = lambda: self.stream_process_unregister(view)
        self.stream_process_register(view)
        return view

    def parameters(self, name=None):
        """Get the list of :class:`joulescope.parameter.Parameter` instances.

        :param name: The optional name of the parameter to retrieve.
            None (default) returns a list of all parameters.
        :return: The list of all parameters.  If name is provided, then just
            return that single parameters.
        """
        if name is not None:
            for p in PARAMETERS:
                if p.name == name:
                    return copy.deepcopy(p)
            return None
        return copy.deepcopy(PARAMETERS)

    def parameter_set(self, name, value):
        """Set a parameter value.

        :param name: The parameter name
        :param value: The new parameter value
        :raise KeyError: if name not found.
        :raise ValueError: if value is not allowed
        """
        raise NotImplementedError()

    def parameter_get(self, name, dtype=None):
        """Get a parameter value.

        :param name: The parameter name.
        :param dtype: The data type for the parameter.  None (default)
            attempts to convert the value to the enum string.
            'actual' returns the value in its actual type used by the driver.
        :raise KeyError: if name not found.
        """
        if name == 'current_ranging':
            pnames = ['type', 'samples_pre', 'samples_window', 'samples_post']
            values = [str(self.parameter_get('current_ranging_' + p)) for p in pnames]
            return '_'.join(values)
        p = PARAMETERS_DICT[name]
        if p.path == 'info':
            return self._parameter_get_info(name)
        value = self._parameters[name]
        if dtype == 'actual':
            return value
        try:
            return value_to_name(name, value)
        except KeyError:
            return value

    def _topic_make(self, topic):
        if topic[0] != '/':
            topic = '/' + topic
        return self._path + topic

    def publish(self, topic, value, timeout=None):
        """Publish to the underlying joulescope_driver instance.

        :param topic: The publish topic.
        :param value: The publish value.
        :param timeout: The timeout in float seconds to wait for this operation
            to complete.  None waits the default amount.
            0 does not wait and subscription will occur asynchronously.
        """
        return self._driver.publish(self._topic_make(topic), value, timeout)

    def query(self, topic, timeout=None):
        """Query the underlying joulescope_driver instance.

        :param topic: The publish topic to query.
        :param timeout: The timeout in float seconds to wait for this operation
            to complete.  None waits the default amount.
            0 does not wait and subscription will occur asynchronously.
        :return: The value associated with topic.
        """
        return self._driver.query(self._topic_make(topic), timeout)

    def subscribe(self, topic, flags, fn, timeout=None):
        """Subscribe to receive topic updates.

        :param self: The driver instance.
        :param topic: Subscribe to this topic string.
        :param flags: The flags or list of flags for this subscription.
            The flags can be int32 jsdrv_subscribe_flag_e or string
            mnemonics, which are:

            - pub: Subscribe to normal values
            - pub_retain: Subscribe to normal values and immediately publish
              all matching retained values.  With timeout, this function does
              not return successfully until all retained values have been
              published.
            - metadata_req: Subscribe to metadata requests (not normally useful).
            - metadata_rsp: Subscribe to metadata updates.
            - metadata_rsp_retain: Subscribe to metadata updates and immediately
              publish all matching retained metadata values.
            - query_req: Subscribe to all query requests (not normally useful).
            - query_rsp: Subscribe to all query responses.
            - return_code: Subscribe to all return code responses.

        :param fn: The function to call on each publish.  Note that python
            dynamically constructs bound methods.  To unsubscribe a method,
            provide the exact same bound method instance to unsubscribe.
            This constrain usually means that the caller needs to hold onto
            the instance.method value passed to this function.
        :param timeout: The timeout in float seconds to wait for this operation
            to complete.  None waits the default amount.
            0 does not wait and subscription will occur asynchronously.
        :raise RuntimeError: on subscribe failure.
        """
        return self._driver.subscribe(self._topic_make(topic), flags, fn, timeout)

    def unsubscribe(self, topic, fn, timeout=None):
        """Unsubscribe a callback to a topic.

        :param topic: Unsubscribe from this topic string.
        :param fn: The callback function to unsubscribe.
        :param timeout: The timeout in float seconds to wait for this operation
            to complete.  None waits the default amount.
            0 does not wait and subscription will occur asynchronously.
        """
        return self._driver.unsubscribe(self._topic_make(topic), fn, timeout)

    def unsubscribe_all(self, fn, timeout=None):
        """Unsubscribe a callback from all topics.

        :param fn: The callback function to unsubscribe.
        :param timeout: The timeout in float seconds to wait for this operation
            to complete.  None waits the default amount.
            0 does not wait and subscription will occur asynchronously.
        """
        return self._driver.unsubscribe(fn, timeout)

    def _config_apply(self, config=None):
        """Apply a configuration set by scan.

        :param config: The configuration string.
        """
        pass

    def open(self, event_callback_fn=None, mode=None, timeout=None):
        """Open this device.

        :param event_callback_fn: The function(event, message) to call on
            asynchronous events, mostly to allow robust handling of device
            errors.  "event" is one of the :class:`DeviceEvent` values,
            and the message is a more detailed description of the event.
        :param mode: The open mode which is one of:
            * 'defaults': Reconfigure the device for default operation.
            * 'restore': Update our state with the current device state.
            * 'raw': Open the device in raw mode for development or firmware update.
            * None: equivalent to 'defaults'.
        :param timeout: The timeout in seconds.  None uses the default timeout.
        """
        rc = self._driver.open(self._path, mode, timeout)
        self.is_open = True
        self.publish('h/fs', 2000000)
        while len(self._parameter_set_queue):
            name, value = self._parameter_set_queue.pop(0)
            self.parameter_set(name, value)
        if len(self._statistics_callbacks):
            self._statistics_start()
        device = 'js110' if 'js110' in self._path.lower() else 'js220'
        self.stream_buffer = StreamBuffer(self._buffer_duration,
                                          frequency=self._input_sampling_frequency,
                                          device=device,
                                          output_frequency=self._output_sampling_frequency)
        self._config_apply(self.config)
        return rc

    def close(self, timeout=None):
        """Close this device and release resources.

        :param timeout: The timeout in seconds.  None uses the default timeout.
        """
        if len(self._statistics_callbacks):
            self._statistics_stop()
        self.stop()
        self.is_open = False
        self.stream_buffer = None
        return self._driver.close(self._path, timeout)

    @property
    def model(self):
        return self._path.split('/')[1]

    @property
    def serial_number(self):
        return self._path.split('/')[-1]

    @property
    def device_serial_number(self):
        return self.serial_number

    @property
    def calibration(self):
        return None

    def info(self):
        """Get the device information structure.

        :return: The device information structure.
        """
        raise NotImplementedError()

    def _stream_process_call(self, method, *args, **kwargs):
        rv = False
        b, self._stream_cbk_objs, self._stream_cbk_objs_add = self._stream_cbk_objs + self._stream_cbk_objs_add, [], []
        for obj in b:
            fn = getattr(obj, method, None)
            if not callable(fn):
                self._stream_cbk_objs.append(obj)
                continue
            if obj.driver_active:
                try:
                    rv |= bool(fn(*args, **kwargs))
                    self._stream_cbk_objs.append(obj)
                except Exception:
                    self._log.exception('%s %s() exception', obj, method)
                    obj.driver_active = False
            if not obj.driver_active:
                try:
                    if hasattr(obj, 'close'):
                        obj.close()
                except Exception:
                    self._log.exception('%s close() exception', obj)
        return rv

    def _on_stream(self, topic, value):
        b = self.stream_buffer
        _, e1 = b.sample_id_range
        b.insert(topic, value)
        e0, e2 = b.sample_id_range

        if e1 == e2:
            return False
        if e0 == e2:
            return False
        rv = self._stream_process_call('stream_notify', self.stream_buffer)
        if rv:
            self.stop()
        if self.stream_buffer.is_duration_max or self.stream_buffer.is_contiguous_duration_max:
            self.stop()

    def start(self, stop_fn=None, duration=None, contiguous_duration=None):
        """Start data streaming.

        :param stop_fn: The function(event, message) called when the device stops.
            The device can stop "automatically" on errors.
            Call :meth:`stop` to stop from the caller.
            This function will be called from the USB processing thread.
            Any calls back into self MUST BE resynchronized.
        :param duration: The duration in seconds for the capture.
        :param contiguous_duration: The contiguous duration in seconds for
            the capture.  As opposed to duration, this ensures that the
            duration has no missing samples.  Missing samples usually
            occur when the device first starts.

        If streaming was already in progress, it will be restarted.
        """
        self.stop()
        self.stream_buffer.reset()
        self.stream_buffer.duration_max = duration
        self.stream_buffer.contiguous_duration_max = contiguous_duration
        self._stop_fn = stop_fn
        for topic, b in zip(self._stream_topics, self.stream_buffer.buffers.values()):
            if topic is None:
                b.active = False
                continue
            b.active = True
            self.subscribe(topic + '!data', 'pub', self._on_stream_cbk)
            self.publish(topic + 'ctrl', 1)
        self._is_streaming = True
        self._stream_process_call('start', self.stream_buffer)

    def stop(self):
        """Stop data streaming.

        :return: True if stopped.  False if was already stopped.

        This method is always safe to call, even after the device has been
        stopped or removed.
        """
        if self._is_streaming:
            self._is_streaming = False
            for topic in self._stream_topics:
                if topic is not None:
                    self.unsubscribe(topic + '!data', self._on_stream_cbk, timeout=0)
                    self.publish(topic + 'ctrl', 0, timeout=0)
            fn, self._stop_fn = self._stop_fn, None
            if callable(fn):
                fn(0, '')  # status, message
            self._stream_process_call('stop')

    def read(self, duration=None, contiguous_duration=None, out_format=None, fields=None):
        """Read data from the device.

        :param duration: The duration in seconds for the capture.
            The duration must fit within the stream_buffer.
        :param contiguous_duration: The contiguous duration in seconds for
            the capture.  As opposed to duration, this ensures that the
            duration has no missing samples.  Missing samples usually
            occur when the device first starts.
            The duration must fit within the stream_buffer.
        :param out_format: The output format which is one of:

            * calibrated: The Nx2 np.ndarray(float32) with columns current and voltage.
            * samples_get: The StreamBuffer samples get format.  Use the fields
              parameter to optionally specify the signals to include.
            * None: equivalent to 'calibrated'.

        :param fields: The fields for samples_get when out_format=samples_get.

        If streaming was already in progress, it will be restarted.
        If neither duration or contiguous duration is specified, the capture
        will only be stopped by callbacks registered through
        :meth:`stream_process_register`.
        """
        self._log.info('read(duration=%s, contiguous_duration=%s, out_format=%s)',
                 duration, contiguous_duration, out_format)
        if out_format not in ['calibrated', 'samples_get', None]:
            raise ValueError(f'Invalid out_format {out_format}')
        if duration is None and contiguous_duration is None:
            raise ValueError('Must specify duration or contiguous_duration')
        duration_max = len(self.stream_buffer) / self._output_sampling_frequency
        if contiguous_duration is not None and contiguous_duration > duration_max:
            raise ValueError(f'contiguous_duration {contiguous_duration} > {duration_max} max seconds')
        if duration is not None and duration > duration_max:
            raise ValueError(f'duration {duration} > {duration_max} max seconds')
        q = queue.Queue()

        def on_stop(*args, **kwargs):
            self._log.info('received stop callback: pending stop')
            q.put(None)

        self.start(on_stop, duration=duration, contiguous_duration=contiguous_duration)
        q.get()
        self.stop()
        start_id, end_id = self.stream_buffer.sample_id_range
        self._log.info('read available range %s, %s', start_id, end_id)
        if contiguous_duration is not None:
            start_id = end_id - int(contiguous_duration * self._output_sampling_frequency)
        elif duration is not None:
            start_id = end_id - int(duration * self._output_sampling_frequency)
        if start_id < 0:
            start_id = 0
        self._log.info('read actual %s, %s', start_id, end_id)

        if out_format in ['calibrated', None]:
            data = self.stream_buffer.samples_get(start_id, end_id, fields=['current', 'voltage'])
            i = data['signals']['current']['value']
            v = data['signals']['voltage']['value']
            return np.hstack([np.reshape(i, (-1, 1)), np.reshape(v, (-1, 1))])
        else:
            return self.stream_buffer.samples_get(start_id, end_id, fields=fields)

    @property
    def is_streaming(self):
        """Check if the device is streaming.

        :return: True if streaming.  False if not streaming.
        """
        return self._is_streaming

    def stream_process_register(self, obj):
        """Register a stream process object.

        :param obj: The instance compatible with :class:`StreamProcessApi`.
            The instance must remain valid until its :meth:`close` is
            called.

        Call :meth:`stream_process_unregister` to disconnect the instance.
        """
        if self._is_streaming and hasattr(obj, 'start'):
            obj.start(self.stream_buffer)
        obj.driver_active = True
        self._stream_cbk_objs_add.append(obj)

    def stream_process_unregister(self, obj):
        """Unregister a stream process object.

        :param obj: The instance compatible with :class:`StreamProcessApi` that was
            previously registered using :meth:`stream_process_register`.
        """
        obj.driver_active = False

    def status(self):
        """Get the current device status.

        :return: A dict containing status information.
        """
        return {
            'driver': {
                'return_code': {
                    'value': 0,
                }
            }
        }

    def extio_status(self):
        """Read the EXTIO GPI value.

        :return: A dict containing the extio status.  Each key is the status
            item name.  The value is itself a dict with the following keys:

            * name: The status name, which is the same as the top-level key.
            * value: The actual value
            * units: The units, if applicable.
            * format: The recommended formatting string (optional).
        """
        return {}

    def __enter__(self):
        """Device context manager, automatically open."""
        self.open()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Device context manager, automatically close."""
        self.close()

    def _query_gpi_value(self):
        gpi_value = None

        def on_gpi_value(topic, value):
            nonlocal gpi_value
            gpi_value = value

        self.subscribe('s/gpi/+/!value', 'pub', on_gpi_value)
        self.publish('s/gpi/+/!req', 0)
        t_start = time.time()
        while gpi_value is None:
            if time.time() - t_start > 1.0:
                raise RuntimeError('_query_gpi_value timed out')
            time.sleep(0.001)
        self.unsubscribe('s/gpi/+/!value', on_gpi_value)
        return gpi_value

    def extio_status(self):
        """Read the EXTIO GPI value.

        :return: A dict containing the extio status.  Each key is the status
            item name.  The value is itself a dict with the following keys:

            * name: The status name, which is the same as the top-level key.
            * value: The actual value
            * units: The units, if applicable.
            * format: The recommended formatting string (optional).
            
        The most interesting key is "gpi_value" which returns the present
        general purpose input signal values.  The remaining keys simply
        copy parameter settings for convenience.
        """
        gpi_value = self._query_gpi_value()
        status = {
            'flags': {
                'value': 0,
                'units': ''},
            'trigger_source': {
                'value': self.parameter_get('trigger_source'),
                'units': ''},
            'current_lsb': {
                'value': self.parameter_get('current_lsb'),
                'units': ''},
            'voltage_lsb': {
                'value': self.parameter_get('voltage_lsb'),
                'units': ''},
            'gpo0': {
                'value': self.parameter_get('gpo0'),
                'units': ''},
            'gpo1': {
                'value': self.parameter_get('gpo1'),
                'units': ''},
            'gpi_value': {
                'value': gpi_value,
                'units': '',
            },
            'io_voltage': {
                'value': self.parameter_get('io_voltage'),
                'units': 'mV',
            },
        }
        for key, value in status.items():
            value['name'] = key
        return status



# Copyright 2019 Jetperch LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
from joulescope.v0.pattern_buffer import PatternBuffer
import numpy as np


def construct(count, counter_first=None, sample_drop=None):
    packet_buffer = np.zeros(count * 512, dtype=np.uint8)
    pkt = packet_buffer.view(dtype=np.uint32)
    pkt_count = 0
    sample_id = 0
    bitshift = 0
    counter = 0
    while len(pkt):
        pkt[0] = 0x02000080
        pkt[1] = pkt_count
        pkt_count += 1
        for idx in range(2, 128):
            while True:
                if bool(counter_first) != ((sample_id & 1) == 0):
                    pkt[idx] = ((~counter & 0xffff) << 16) | counter
                    counter = (counter + 1) & 0xffff
                else:
                    pkt[idx] = bitshift
                    if bitshift == 0:
                        bitshift = 1
                    else:
                        bitshift = (bitshift << 1) & 0xffffffff
                sample_id += 1
                if sample_drop is not None and len(sample_drop) and sample_drop[0] == sample_id:
                    sample_drop.pop(0)
                else:
                    break
        pkt = pkt[128:]
    return packet_buffer


class TestPatternBuffer(unittest.TestCase):

    def test_counter_first(self):
        p = PatternBuffer()
        data = construct(1, counter_first=True)
        self.assertEqual(512, len(data))
        p.insert(data)
        s = p.status()
        self.assertEqual(126, s['sample_id'])
        self.assertEqual(0, s['header_error'])
        self.assertEqual(0, s['pkt_index_error'])
        self.assertEqual(0, s['pattern_error'])

    def test_bitshift_first(self):
        p = PatternBuffer()
        data = construct(2, counter_first=False)
        self.assertEqual(1024, len(data))
        p.insert(data)
        s = p.status()
        self.assertEqual(252, s['sample_id'])
        self.assertEqual(0, s['header_error'])
        self.assertEqual(0, s['pkt_index_error'])
        self.assertEqual(0, s['pattern_error'])

    def test_sample_drop(self):
        p = PatternBuffer()
        data = construct(2, sample_drop=[10, 19])
        p.insert(data)
        s = p.status()
        self.assertEqual(252, s['sample_id'])
        self.assertEqual(0, s['header_error'])
        self.assertEqual(0, s['pkt_index_error'])
        self.assertEqual(2, s['pattern_error'])

    def test_packet_index_error(self):
        p = PatternBuffer()
        data = construct(2)
        data[516] = 3
        p.insert(data)
        s = p.status()
        print(s)
        self.assertEqual(252, s['sample_id'])
        self.assertEqual(0, s['header_error'])
        self.assertEqual(1, s['pkt_index_error'])
        self.assertEqual(0, s['pattern_error'])



# Copyright 2018 Jetperch LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class Parameter:
    """Define a single parameter accessible by the application.

    :param name: The parameter name.
    :param path: None if the parameter is only changed at the streaming start.
        'setting' if the parameter is part of the settings packet.
    :param default: The default value.
    :param options: The list of acceptable values as ('name', value, ['alias1', ...]) tuples.
    :param units: The units string for the parameter.
    :param brief: The brief user-meaningful description.
    :param detail: The detailed user-meaningful description.
    :param flags: The list of flags which include:
        * developer: This parameter is intended for developers only, not end users.
        * read_only: This parameter can only be read (no write permission).
    :param validator: Function called to validate the parameter value.
    """
    def __init__(self, name, path, default=None, options=None, units=None,
                 brief=None, detail=None,
                 flags=None, validator=None):
        self.name = name
        self.default = default
        self.path = path
        self.options = []
        self.str_to_value = {}
        self.units = units
        self.brief = brief
        self.detail = detail
        self.flags = [] if flags is None else flags
        self.validator = validator
        if options is None:
            return
        for idx, option in enumerate(options):
            if len(option) < 2 or len(option) > 3:
                raise ValueError('Parameter %s, value %d invalid: %r' % (name, idx, option))
            vname = option[0]
            vvalue = option[1]
            if not isinstance(vname, str):
                raise ValueError('Parameter %s, value %d invalid name %r' % (name, idx, vname))
            if len(option) == 2:
                option = (vname, vvalue, [])
            self.options.append(option)
            self._insert(vname, vvalue)
            for alias in option[2]:
                self._insert(alias, vvalue)

        if default not in [x[0] for x in options]:
            raise ValueError('Parameter %s, default %r not found' % (name, default))

    def __str__(self):
        return f'Parameter({self.name})'

    def _insert(self, key, value):
        if key in self.str_to_value:
            raise ValueError('Parameter %s: key %s already exists' % (self.name, key))
        self.str_to_value[key] = value



# Copyright 2020 Jetperch LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This executable displays information about the joulescope package and
connected devices.
"""

import joulescope
import logging
import platform
import sys


def parser_config(p):
    """Display platform and connected device details."""
    return on_cmd


def on_cmd(args):
    frozen = getattr(sys, 'frozen', False)
    if frozen:
        frozen = getattr(sys, '_MEIPASS', frozen)
    print('System information')
    print(f'    Python: {sys.version}')
    print(f'    Platform: {platform.platform()} ({sys.platform})')
    print(f'    Processor: {platform.processor()}')
    print(f'    executable: {sys.executable}')
    print(f'    frozen: {frozen}')

    print('')
    print(f'joulescope version: {joulescope.__version__}')
    devices = joulescope.scan()
    device_count = len(devices)
    if device_count == 0:
        print('Found 0 connected Joulescopes.')
    elif device_count == 1:
        print('Found 1 connected Joulescope:')
    else:
        print(f'Found {device_count} connected Joulescopes:')
    logging.getLogger().setLevel(logging.WARNING)
    for device in devices:
        try:
            with device:
                info = device.info()
            ctl_fw = info.get('ctl', {}).get('fw', {}).get('ver', '')
            sensor_fw = info.get('sensor', {}).get('fw', {}).get('ver', '')
            info = f'  ctl={ctl_fw:<15}  sensor={sensor_fw}'
        except Exception:
            info = ''
        print(f'    {device} {info}')




# Copyright 2018 Jetperch LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Joulescope uses public key cryptography to sign calibrations and
firmware update.  This file contains the public keys.
"""

from binascii import unhexlify as uhx

CALIBRATION_SIGNING = uhx(b'3b3aee3fb2ac984fd37040aa52092e2ad88371455285002a6eef568ef152232f')
RECALIBRATION_01_SIGNING = uhx(b'75693bf67e3a8d783789e831e16a8df40fee55b363e787d2212866b1a18cf64d')
FIRMWARE_SIGNING = uhx(b'32fe2bed04bbc42fe1b382e0371ba95ec2947045e8d919e49fdef601e24c105e')
FIRMWARE_DISTRIBUTION = uhx(b'83fd0b085bda32c041644b7c299859b1101fbccc334f9ba6223a9d94a083a9d0')
JOULESCOPE_ROOT = uhx(b'eb385a7ff9c1315deba7cd5fca29630a1a0c66b4bbb0deac5c1e18aa68715ae4')



# Copyright 2018 Jetperch LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Test the packet models

"""

import unittest
from joulescope import units
import numpy as np


class TestUnits(unittest.TestCase):

    def assertClose(self, expected, actual):
        np.testing.assert_almost_equal(expected[0], actual[0])
        self.assertEqual(expected[1], actual[1])
        np.testing.assert_almost_equal(expected[2], actual[2])

    def test_no_change(self):
        self.assertClose((0.0, '', 1), units.unit_prefix(0.0))
        self.assertClose((1.0, '', 1), units.unit_prefix(1.0))
        self.assertClose((5.0, '', 1), units.unit_prefix(5.0))

    def test_scaled(self):
        self.assertClose((10.0, 'G', 1e9), units.unit_prefix(10e9))
        self.assertClose((10.0, 'M', 1e6), units.unit_prefix(10e6))
        self.assertClose((10.0, 'k', 1e3), units.unit_prefix(10e3))
        self.assertClose((100.0, 'm', 0.001), units.unit_prefix(0.1))
        self.assertClose((100.0, 'µ', 0.000001), units.unit_prefix(0.0001))
        self.assertClose((10.0, 'n', 1e-9), units.unit_prefix(10e-9))
        self.assertClose((10.0, 'p', 1e-12), units.unit_prefix(10e-12))

    def test_three_sig_figs_no_units(self):
        self.assertEqual('1.50', units.three_sig_figs(1.5))
        self.assertEqual('150m', units.three_sig_figs(0.15))
        self.assertEqual('150µ', units.three_sig_figs(0.00015))
        self.assertEqual('1.50k', units.three_sig_figs(1500))
        self.assertEqual('1.50M', units.three_sig_figs(1500000))

    def test_three_sig_figs_with_units(self):
        self.assertEqual('1.50 A', units.three_sig_figs(1.5, 'A'))
        self.assertEqual('150 mA', units.three_sig_figs(0.15, 'A'))
        self.assertEqual('15.0 mA', units.three_sig_figs(0.015, 'A'))
        self.assertEqual('1.50 mA', units.three_sig_figs(0.0015, 'A'))
        self.assertEqual('150 µA', units.three_sig_figs(0.00015, 'A'))

    def test_three_sig_figs_boundary(self):
        self.assertEqual('1.00 A', units.three_sig_figs(1.0, 'A'))
        self.assertEqual('100 mA', units.three_sig_figs(0.10, 'A'))
        self.assertEqual('10.0 mA', units.three_sig_figs(0.010, 'A'))
        self.assertEqual('1.00 mA', units.three_sig_figs(0.0010, 'A'))
        self.assertEqual('100 µA', units.three_sig_figs(0.00010, 'A'))

    def test_convert_number(self):
        self.assertEqual(1, units.str_to_number('1'))
        self.assertEqual(-1, units.str_to_number('-1'))
        self.assertEqual(1000, units.str_to_number('1000'))
        self.assertEqual(1000, units.str_to_number('1k'))
        self.assertEqual(1000, units.str_to_number('1kOhm'))
        self.assertEqual(1000, units.str_to_number('1 kOhm'))
        self.assertEqual(1e-6, units.str_to_number('1uF'))

    def test_negative_numbers(self):
        self.assertEqual('-1.50 A', units.three_sig_figs(-1.5, 'A'))
        self.assertEqual('-150 mA', units.three_sig_figs(-0.15, 'A'))
        self.assertEqual('-15.0 mA', units.three_sig_figs(-0.015, 'A'))
        self.assertEqual('-1.50 mA', units.three_sig_figs(-0.0015, 'A'))

    def test_duration_to_seconds(self):
        with self.assertRaises(ValueError):
            units.duration_to_seconds(None)
        with self.assertRaises(ValueError):
            units.duration_to_seconds('')
        self.assertEqual(1.0, units.duration_to_seconds(1))
        self.assertEqual(1.0, units.duration_to_seconds(1.0))
        self.assertEqual(1.0, units.duration_to_seconds('1'))
        self.assertEqual(1.0, units.duration_to_seconds('1s'))
        self.assertEqual(1.5, units.duration_to_seconds('1.5'))
        self.assertEqual(1.5, units.duration_to_seconds('1.5s'))
        self.assertEqual(60.0, units.duration_to_seconds('1m'))
        self.assertEqual(90.0, units.duration_to_seconds('1.5m'))
        self.assertEqual(3600.0, units.duration_to_seconds('1h'))
        self.assertEqual(86400.0, units.duration_to_seconds('1d'))
        with self.assertRaises(ValueError):
            units.duration_to_seconds('hello')



# Copyright 2022 Jetperch LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Test the circular sample buffer.
"""

import unittest
from joulescope.v1.sample_buffer import SampleBuffer
import numpy as np


class TestSampleBuffer(unittest.TestCase):

    def test_new(self):
        c = SampleBuffer(10, dtype=np.float32)
        self.assertEqual(0, len(c))
        self.assertEqual((0, 0), c.range)
        with self.assertRaises(ValueError):
            c.get_range(0, 1)

    def test_add_one(self):
        c = SampleBuffer(10, dtype=np.float32)
        a = np.array([2.0], dtype=np.float32)
        c.add(0, a)
        self.assertEqual(1, len(c))
        self.assertEqual((0, 1), c.range)
        np.testing.assert_equal(a, c.get_range(*c.range))
        with self.assertRaises(ValueError):
            c.get_range(-1, 0)
        with self.assertRaises(ValueError):
            c.get_range(1, 2)

    def test_add_one_offset(self):
        c = SampleBuffer(10, dtype=np.float32)
        a = np.array([2.0], dtype=np.float32)
        c.add(111, a)
        self.assertEqual(1, len(c))
        self.assertEqual((111, 112), c.range)
        np.testing.assert_equal(a, c.get_range(*c.range))
        with self.assertRaises(ValueError):
            c.get_range(110, 111)
        with self.assertRaises(ValueError):
            c.get_range(112, 113)

    def test_add_one_overfill(self):
        c = SampleBuffer(10, dtype=np.float32)
        for i in range(20):
            c.add(i, np.array([i], dtype=np.float32))
        self.assertEqual(9, len(c))
        self.assertEqual((11, 20), c.range)
        np.testing.assert_equal(np.arange(11, 20, 1, dtype=np.float32), c.get_range(*c.range))

    def test_add_multiple_overflow_initial(self):
        c = SampleBuffer(10, dtype=np.float32)
        a = np.arange(9, 15, dtype=np.float32)
        c.add(9, a)
        self.assertEqual(len(a), len(c))
        self.assertEqual((9, 15), c.range)
        np.testing.assert_equal(a, c.get_range(*c.range))
        np.testing.assert_equal(a[1:], c.get_range(10, 15))
        np.testing.assert_equal(a[:2], c.get_range(9, 11))

    def test_skip(self):
        c = SampleBuffer(10, dtype=np.float32)
        c.add(1, [1.0])
        c.add(3, [3.0])
        self.assertEqual(3, len(c))
        self.assertEqual((1, 4), c.range)
        s = c.get_range(*c.range)
        self.assertEqual(1.0, s[0])
        self.assertTrue(np.isnan(s[1]))
        self.assertEqual(3.0, s[2])



# Copyright 2018 Jetperch LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# https://msdn.microsoft.com/en-us/library/windows/hardware/ff550897(v=vs.85).aspx
from ctypes import windll, Structure, sizeof, resize, byref, wstring_at, \
    c_byte, c_ulong, c_wchar_p, c_void_p, POINTER
from ctypes.wintypes import DWORD, WORD, WCHAR, HANDLE, BOOL
import uuid

_setupapi = windll.SetupApi


# SetupDiGetClassDevs flags
DIGCF_DEFAULT = 0x00000001
DIGCF_PRESENT = 0x00000002
DIGCF_ALLCLASSES = 0x00000004
DIGCF_PROFILE = 0x00000008
DIGCF_DEVICE_INTERFACE = 0x00000010


class GUID(Structure):
    _fields_ = [
        ('data1', DWORD), 
        ('data2', WORD),
        ('data3', WORD), 
        ('data4', c_byte * 8),
    ]
    
    def __init__(self, guid=None):
        Structure.__init__(self)
        if guid:
            x = uuid.UUID(guid)
            f = x.fields
            self.data1 = DWORD(f[0])
            self.data2 = WORD(f[1])
            self.data3 = WORD(f[2])
            data4_cast = c_byte * 8
            self.data4 = data4_cast(*x.bytes[8:])


class SpDevinfoData(Structure):
    _fields_ = [
        ('cb_size', DWORD), 
        ('class_guid', GUID),
        ('dev_inst', DWORD), 
        ('reserved', POINTER(c_ulong)),
    ]

    def __init__(self):
        Structure.__init__(self)
        self.cb_size = sizeof(self)


class SpDeviceInterfaceData(Structure):
    _fields_ = [
        ('cb_size', DWORD), 
        ('interface_class_guid', GUID),
        ('flags', DWORD), 
        ('reserved', POINTER(c_ulong)),
    ]
    
    def __init__(self):
        Structure.__init__(self)
        self.cb_size = sizeof(self)


class SpDeviceInterfaceDetailData(Structure):
    _fields_ = [
        ("cb_size", DWORD), 
        ("device_path", WCHAR * 1),  # variable size character array
    ]

    def __init__(self):
        Structure.__init__(self)
        self.cb_size = sizeof(self)
    

# HDEVINFO SetupDiGetClassDevs(
#   _In_opt_ const GUID   *ClassGuid,
#   _In_opt_       PCTSTR Enumerator,
#   _In_opt_       HWND   hwndParent,
#   _In_           DWORD  Flags);
SetupDiGetClassDevs = _setupapi.SetupDiGetClassDevsW
SetupDiGetClassDevs.restype = HANDLE
SetupDiGetClassDevs.argtypes = [POINTER(GUID), c_wchar_p, HANDLE, DWORD]

# BOOL SetupDiEnumDeviceInterfaces(
#   _In_           HDEVINFO                  DeviceInfoSet,
#   _In_opt_       PSP_DEVINFO_DATA          DeviceInfoData,
#   _In_     const GUID                      *InterfaceClassGuid,
#   _In_           DWORD                     MemberIndex,
#   _Out_          PSP_DEVICE_INTERFACE_DATA DeviceInterfaceData);
SetupDiEnumDeviceInterfaces = _setupapi.SetupDiEnumDeviceInterfaces
SetupDiEnumDeviceInterfaces.restype = BOOL
SetupDiEnumDeviceInterfaces.argtypes = [c_void_p, POINTER(SpDevinfoData), POINTER(GUID), DWORD, POINTER(SpDeviceInterfaceData)]

# BOOL SetupDiGetDeviceInterfaceDetail(
#   _In_      HDEVINFO                         DeviceInfoSet,
#   _In_      PSP_DEVICE_INTERFACE_DATA        DeviceInterfaceData,
#   _Out_opt_ PSP_DEVICE_INTERFACE_DETAIL_DATA DeviceInterfaceDetailData,
#   _In_      DWORD                            DeviceInterfaceDetailDataSize,
#   _Out_opt_ PDWORD                           RequiredSize,
#   _Out_opt_ PSP_DEVINFO_DATA                 DeviceInfoData);
SetupDiGetDeviceInterfaceDetail = _setupapi.SetupDiGetDeviceInterfaceDetailW
SetupDiGetDeviceInterfaceDetail.restype = BOOL
SetupDiGetDeviceInterfaceDetail.argtypes = [c_void_p, POINTER(SpDeviceInterfaceData), POINTER(SpDeviceInterfaceDetailData), DWORD, POINTER(DWORD), POINTER(SpDevinfoData)]

# BOOL SetupDiDestroyDeviceInfoList(
#   _In_ HDEVINFO DeviceInfoSet);
SetupDiDestroyDeviceInfoList = _setupapi.SetupDiDestroyDeviceInfoList
SetupDiDestroyDeviceInfoList.restype = BOOL
SetupDiDestroyDeviceInfoList.argtypes = [c_void_p]


def device_interface_guid_to_paths(guid):
    """Find the connected devices that match the provided GUID.
    
    :param guid: The deviceInterfaceGUID to match.
    :return: The list of path strings.
    """
    paths = []
    flags = DWORD(DIGCF_PRESENT | DIGCF_DEVICE_INTERFACE)
    guid = GUID(guid)
    handle = SetupDiGetClassDevs(byref(guid), None, None, flags)
    if handle is None:
        raise RuntimeError('SetupDiGetClassDevs invalid')

    sp_device_interface_data = SpDeviceInterfaceData()
    sp_device_info_data = SpDevinfoData()
    member_index = 0

    while SetupDiEnumDeviceInterfaces(handle, None, byref(guid), DWORD(member_index), byref(sp_device_interface_data)):
        required_size = c_ulong(0)
        bResult = SetupDiGetDeviceInterfaceDetail(handle, byref(sp_device_interface_data), None, 0, byref(required_size), None)
        sp_device_interface_detail_data = SpDeviceInterfaceDetailData()
        resize(sp_device_interface_detail_data, required_size.value)
        sp_device_interface_detail_data.cb_size = sizeof(SpDeviceInterfaceDetailData)
        bResult = SetupDiGetDeviceInterfaceDetail(
            handle, byref(sp_device_interface_data), 
            byref(sp_device_interface_detail_data), required_size, 
            byref(required_size), byref(sp_device_info_data))
        if bResult:
            path = wstring_at(byref(sp_device_interface_detail_data, sizeof(DWORD)))
            paths.append(path)
        member_index += 1
    SetupDiDestroyDeviceInfoList(handle)
    return paths
